﻿--[[
Name: LibDogTag-2.0
Revision: $Rev: 59953 $
Author: Cameron Kenneth Knight (ckknight@gmail.com)
Inspired By: WatchDog by Vika
Website: http://www.wowace.com/
Documentation: http://www.wowace.com/wiki/LibDogTag-2.0
SVN: svn://svn.wowace.com/wowace/trunk/LibDogTag-2.0
Description: A library to provide a markup syntax for unit frame FontStrings
Dependencies: Rock (optional), MobHealth3/MobHealth2/MobInfo2 (optional), Babble-Spell-2.2 (optional), Threat-1.0 (optional), RangeCheck-1.0 (optional)
License: LGPL v2.1
]]

-------------------------------------------------------------------
-- FOR TAG INFORMATION SEE http://www.wowace.com/wiki/DogTag-2.0 --
-------------------------------------------------------------------

local MAJOR_VERSION = "LibDogTag-2.0"
local MINOR_VERSION = tonumber(("$Revision: 59953 $"):match("%d+")) or 0

_G.DogTag_MINOR_VERSION = MINOR_VERSION

_G.DogTag_funcs = {}

DogTag_funcs[#DogTag_funcs+1] = function()

local DogTag = _G.DogTag

local oldLib
if next(DogTag) ~= nil then
	oldLib = {}
	for k,v in pairs(DogTag) do
		oldLib[k] = v
		DogTag[k] = nil
	end
end

local L = {
	["DogTag Help"] = "DogTag Help",
	
	-- races
	["Blood Elf"] = "Blood Elf",
	["Draenei"] = "Draenei",
	["Dwarf"] = "Dwarf",
	["Gnome"] = "Gnome",
	["Human"] = "Human",
	["Night Elf"] = "Night Elf",
	["Orc"] = "Orc",
	["Tauren"] = "Tauren",
	["Troll"] = "Troll",
	["Undead"] = "Undead",

	-- short races
	["Blood Elf_short"] = "BE",
	["Draenei_short"] = "Dr",
	["Dwarf_short"] = "Dw",
	["Gnome_short"] = "Gn",
	["Human_short"] = "Hu",
	["Night Elf_short"] = "NE",
	["Orc_short"] = "Or",
	["Tauren_short"] = "Ta",
	["Troll_short"] = "Tr",
	["Undead_short"] = "Ud",

	-- classes
	["Warrior"] = "Warrior",
	["Priest"] = "Priest",
	["Mage"] = "Mage",
	["Shaman"] = "Shaman",
	["Paladin"] = "Paladin",
	["Warlock"] = "Warlock",
	["Druid"] = "Druid",
	["Rogue"] = "Rogue",
	["Hunter"] = "Hunter",

	-- short classes
	["Warrior_short"] = "Wr",
	["Priest_short"] = "Pr",
	["Mage_short"] = "Ma",
	["Shaman_short"] = "Sh",
	["Paladin_short"] = "Pa",
	["Warlock_short"] = "Wl",
	["Druid_short"] = "Dr",
	["Rogue_short"] = "Ro",
	["Hunter_short"] = "Hu",

	["Player"] = PLAYER,
	["Target"] = TARGET,
	["Focus-target"] = FOCUS,
	["Mouse-over"] = "Mouse-over",
	["%s's pet"] = "%s's pet",
	["%s's target"] = "%s's target",
	["Party member #%d"] = "Party member #%d",
	["Raid member #%d"] = "Raid member #%d",

	-- classifications
	["Rare"] = ITEM_QUALITY3_DESC,
	["Rare-Elite"] = ITEM_QUALITY3_DESC .. "-" .. ELITE,
	["Elite"] = ELITE,
	["Boss"] = BOSS,
	-- short classifications
	["Rare_short"] = "r",
	["Rare-Elite_short"] = "r+",
	["Elite_short"] = "+",
	["Boss_short"] = "b",

	["Feigned Death"] = "Feigned Death",
	["Stealthed"] = "Stealthed",
	["Soulstoned"] = "Soulstoned",

	["Dead"] = DEAD,
	["Ghost"] = "Ghost",
	["Offline"] = PLAYER_OFFLINE,
	["Online"] = "Online",
	["Combat"] = "Combat",
	["Resting"] = "Resting",
	["Tapped"] = "Tapped",
	["AFK"] = "AFK",
	["DND"] = "DND",

	["True"] = "True",

	["Rage"] = RAGE,
	["Focus"] = FOCUS,
	["Energy"] = ENERGY,
	["Mana"] = MANA,

	["PvP"] = PVP,
	["FFA"] = "FFA",

	-- genders
	["Male"] = MALE,
	["Female"] = FEMALE,

	-- forms
	["Bear"] = "Bear",
	["Cat"] = "Cat",
	["Moonkin"] = "Moonkin",
	["Aquatic"] = "Aquatic",
	["Flight"] = "Flight",
	["Travel"] = "Travel",
	["Tree"] = "Tree",

	["Bear_short"] = "Be",
	["Cat_short"] = "Ca",
	["Moonkin_short"] = "Mk",
	["Aquatic_short"] = "Aq",
	["Flight_short"] = "Fl",
	["Travel_short"] = "Tv",
	["Tree_short"] = "Tr",

	-- shortgenders
	["Male_short"] = "m",
	["Female_short"] = "f",

	["Leader"] = "Leader",

	-- spell trees
	["Hybrid"] = "Hybrid", -- for all 3 trees
	["Druid_Tree_1"] = "Balance",
	["Druid_Tree_2"] = "Feral Combat",
	["Druid_Tree_3"] = "Restoration",
	["Hunter_Tree_1"] = "Beast Mastery",
	["Hunter_Tree_2"] = "Marksmanship",
	["Hunter_Tree_3"] = "Survival",
	["Mage_Tree_1"] = "Arcane",
	["Mage_Tree_2"] = "Fire",
	["Mage_Tree_3"] = "Frost",
	["Paladin_Tree_1"] = "Holy",
	["Paladin_Tree_2"] = "Protection",
	["Paladin_Tree_3"] = "Retribution",
	["Priest_Tree_1"] = "Discipline",
	["Priest_Tree_2"] = "Holy",
	["Priest_Tree_3"] = "Shadow",
	["Rogue_Tree_1"] = "Assassination",
	["Rogue_Tree_2"] = "Combat",
	["Rogue_Tree_3"] = "Subtlety",
	["Shaman_Tree_1"] = "Elemental",
	["Shaman_Tree_2"] = "Enhancement",
	["Shaman_Tree_3"] = "Restoration",
	["Warrior_Tree_1"] = "Arms",
	["Warrior_Tree_2"] = "Fury",
	["Warrior_Tree_3"] = "Protection",
	["Warlock_Tree_1"] = "Affliction",
	["Warlock_Tree_2"] = "Demonology",
	["Warlock_Tree_3"] = "Destruction",
}
for k,v in pairs(L) do
	if type(v) ~= "string" then -- some evil addon messed it up
		L[k] = k
	end
end

-- alter L here
if GetLocale() == "deDE" then
	-- races
	L["Blood Elf"] = "Blutelf"
	L["Dwarf"] = "Zwerg"
	L["Gnome"] = "Gnom"
	L["Human"] = "Mensch"
	L["Night Elf"] = "Nachtelf"
	L["Undead"] = "Untoter"

	-- classes
	L["Warrior"] = "Krieger"
	L["Priest"] = "Priester"
	L["Mage"] = "Magier"
	L["Shaman"] = "Schamane"
	L["Warlock"] = "Hexenmeister"
	L["Druid"] = "Druide"
	L["Rogue"] = "Schurke"
	L["Hunter"] = "J\195\164ger"

	-- short classes
	L["Warrior_short"] = "Kr"
	L["Warlock_short"] = "He"
	L["Hunter_short"] = "J\195\164"

	L["%s's pet"] = "%s's Begleiter"
	L["%s's target"] = "%s's Ziel"
	L["Party member #%d"] = "Party Mitglied #%d"
	L["Raid member #%d"] = "Schlachtzug #%d"

	L["Feigned Death"] = "Totgestellt"
	L["Stealthed"] = "Verstohlen"
	L["Soulstoned"] = "Seelenstein"

	L["Ghost"] = "Geist"
	L["Online"] = "Online"
	L["Combat"] = "Kampf"
	L["Resting"] = "Ausgeruht"
	L["Tapped"] = "angezapft"

	-- forms
	L["Bear"] = "B\195\164r"
	L["Cat"] = "Katze"
	L["Moonkin"] = "Mondkin"
	L["Aquatic"] = "Wasser"
	L["Flight"] = "Flug"
	L["Travel"] = "Reise"
	L["Tree"] = "Baum"

	L["Bear_short"] = "B\195\164"
	L["Cat_short"] = "Ka"
	L["Moonkin_short"] = "Mk"
	L["Aquatic_short"] = "Wa"
	L["Flight_short"] = "Fl"
	L["Travel_short"] = "Re"
	L["Tree_short"] = "Ba"

	-- shortgenders
	L["Male_short"] = "m"
	L["Female_short"] = "w"

	L["Leader"] = "Leiter"

	-- spell trees
	L["Hybrid"] = "Hybride"
	L["Druid_Tree_1"] = "Gleichgewicht"
	L["Druid_Tree_2"] = "Wilder Kampf"
	L["Druid_Tree_3"] = "Wiederherstellung"
	L["Hunter_Tree_1"] = "Tierherrschaft"
	L["Hunter_Tree_2"] = "Treffsicherheit"
	L["Hunter_Tree_3"] = "Überleben"
	L["Mage_Tree_1"] = "Arkan"
	L["Mage_Tree_2"] = "Feuer"
	L["Paladin_Tree_1"] = "Heilig"
	L["Paladin_Tree_2"] = "Schutz"
	L["Paladin_Tree_3"] = "Vergeltung"
	L["Priest_Tree_1"] = "Disziplin"
	L["Priest_Tree_2"] = "Heilig"
	L["Priest_Tree_3"] = "Schatten"
	L["Rogue_Tree_1"] = "Meucheln"
	L["Rogue_Tree_2"] = "Kampf"
	L["Rogue_Tree_3"] = "T\195\164uschung"
	L["Shaman_Tree_1"] = "Elementar"
	L["Shaman_Tree_2"] = "Verst\195\164rkung"
	L["Shaman_Tree_3"] = "Wiederherstellung"
	L["Warrior_Tree_1"] = "Waffen"
	L["Warrior_Tree_2"] = "Furor"
	L["Warrior_Tree_3"] = "Schutz"
	L["Warlock_Tree_1"] = "Gebrechen"
	L["Warlock_Tree_2"] = "D\195\164monologie"
	L["Warlock_Tree_3"] = "Zerstörung"

elseif GetLocale() == "koKR" then
	-- races
	L["Blood Elf"] = "블러드 엘프"
	L["Draenei"] = "드레나이"
	L["Dwarf"] = "드워프"
	L["Gnome"] = "노움"
	L["Human"] = "인간"
	L["Night Elf"] = "나이트 엘프"
	L["Orc"] = "오크"
	L["Tauren"] = "타우렌"
	L["Troll"] = "트롤"
	L["Undead"] = "언데드"

	-- short races
	L["Blood Elf_short"] = "블엘"
	L["Draenei_short"] = "드레"
	L["Dwarf_short"] = "드웝"
	L["Gnome_short"] = "놈"
	L["Human_short"] = "인간"
	L["Night Elf_short"] = "나엘"
	L["Orc_short"] = "오크"
	L["Tauren_short"] = "타렌"
	L["Troll_short"] = "트롤"
	L["Undead_short"] = "언데"

	-- classes
	L["Warrior"] = "전사"
	L["Priest"] = "사제"
	L["Mage"] = "마법사"
	L["Shaman"] = "주술사"
	L["Paladin"] = "성기사"
	L["Warlock"] = "흑마법사"
	L["Druid"] = "드루이드"
	L["Rogue"] = "도적"
	L["Hunter"] = "사냥꾼"

	-- short classes
	L["Warrior_short"] = "전"
	L["Priest_short"] = "사"
	L["Mage_short"] = "마"
	L["Shaman_short"] = "주"
	L["Paladin_short"] = "성"
	L["Warlock_short"] = "흑"
	L["Druid_short"] = "드"
	L["Rogue_short"] = "도"
	L["Hunter_short"] = "냥"

	L["Focus-target"] = "주시 대상"
	L["Mouse-over"] = "마우스-오버"
	L["%s's pet"] = "%s의 소환수"
	L["%s's target"] = "%s의 대상"
	L["Party member #%d"] = "파티원 #%d"
	L["Raid member #%d"] = "공격대원 #%d"

	L["Rare_short"] = "희"
	L["Rare-Elite_short"] = "희+"
	L["Elite_short"] = "+"
	L["Boss_short"] = "보"

	L["Feigned Death"] = "죽은척하기" -- must match aura
	L["Stealthed"] = "은신중"
	L["Soulstoned"] = "영혼석 보관"

	L["Ghost"] = "유령"
	L["Online"] = "접속중"
	L["Combat"] = "전투중"
	L["Resting"] = "휴식중"
	L["Tapped"] = "선점"
	L["AFK"] = "자리비움"
	L["DND"] = "다른용무중"

	L["True"] = "True" -- check

	L["PvP"] = "전쟁" -- PVP,
	L["FFA"] = "전투 지역"

	-- forms
	L["Bear"] = "곰"
	L["Cat"] = "표범"
	L["Moonkin"] = "달빛야수"
	L["Aquatic"] = "바다표범"
	L["Flight"] = "폭풍까마귀"
	L["Travel"] = "치타"
	L["Tree"] = "나무"

	L["Bear_short"] = "곰"
	L["Cat_short"] = "표범"
	L["Moonkin_short"] = "달빛"
	L["Aquatic_short"] = "바표"
	L["Flight_short"] = "폭까"
	L["Travel_short"] = "치타"
	L["Tree_short"] = "나무"

	-- shortgenders
	L["Male_short"] = "남"
	L["Female_short"] = "여"
	L["Leader"] = "지휘관"

	-- spell trees
	L["Hybrid"] = "하이브리드" -- for all 3 trees
	L["Druid_Tree_1"] = "조화"
	L["Druid_Tree_2"] = "야성"
	L["Druid_Tree_3"] = "회복"
	L["Hunter_Tree_1"] = "야수"
	L["Hunter_Tree_2"] = "사격"
	L["Hunter_Tree_3"] = "생존"
	L["Mage_Tree_1"] = "비전"
	L["Mage_Tree_2"] = "화염"
	L["Mage_Tree_3"] = "냉기"
	L["Paladin_Tree_1"] = "신성"
	L["Paladin_Tree_2"] = "보호"
	L["Paladin_Tree_3"] = "징벌"
	L["Priest_Tree_1"] = "수양"
	L["Priest_Tree_2"] = "신성"
	L["Priest_Tree_3"] = "암흑"
	L["Rogue_Tree_1"] = "암살"
	L["Rogue_Tree_2"] = "전투"
	L["Rogue_Tree_3"] = "잠행"
	L["Shaman_Tree_1"] = "정기"
	L["Shaman_Tree_2"] = "고양"
	L["Shaman_Tree_3"] = "복원"
	L["Warrior_Tree_1"] = "무기"
	L["Warrior_Tree_2"] = "분노"
	L["Warrior_Tree_3"] = "방어"
	L["Warlock_Tree_1"] = "고통"
	L["Warlock_Tree_2"] = "악마"
	L["Warlock_Tree_3"] = "파괴"
elseif GetLocale() == "frFR" then
	-- races
	L["Blood Elf"] = "Elfe de sang"
	L["Draenei"] = "Draeneï"
	L["Dwarf"] = "Nain"
	L["Gnome"] = "Gnome"
	L["Human"] = "Humain"
	L["Night Elf"] = "Elfe de la nuit"
	L["Orc"] = "Orc"
	L["Tauren"] = "Tauren"
	L["Troll"] = "Troll"
	L["Undead"] = "Mort-vivant"

	-- short races
	L["Blood Elf_short"] = "ES"
	L["Draenei_short"] = "Dr"
	L["Dwarf_short"] = "Na"
	L["Gnome_short"] = "Gn"
	L["Human_short"] = "Hu"
	L["Night Elf_short"] = "EN"
	L["Orc_short"] = "Or"
	L["Tauren_short"] = "Ta"
	L["Troll_short"] = "Tr"
	L["Undead_short"] = "MV"

	-- classes
	L["Warrior"] = "Guerrier"
	L["Priest"] = "Prêtre"
	L["Mage"] = "Mage"
	L["Shaman"] = "Chaman"
	L["Paladin"] = "Paladin"
	L["Warlock"] = "Démoniste"
	L["Druid"] = "Druide"
	L["Rogue"] = "Voleur"
	L["Hunter"] = "Chasseur"

	-- short classes
	L["Warrior_short"] = "Gu"
	L["Priest_short"] = "Pr"
	L["Mage_short"] = "Ma"
	L["Shaman_short"] = "Ch"
	L["Paladin_short"] = "Pa"
	L["Warlock_short"] = "Dé"
	L["Druid_short"] = "Dr"
	L["Rogue_short"] = "Vo"
	L["Hunter_short"] = "Ch"

	L["Player"] = PLAYER
	L["Target"] = TARGET
	L["Focus-target"] = FOCUS
	L["Mouse-over"] = "Sous la souris"
	L["%s's pet"] = "Familier de %s"
	L["%s's target"] = "Cible de %s"
	L["Party member #%d"] = "Membre du groupe #%d"
	L["Raid member #%d"] = "Membre du raid #%d"

	-- classifications
	L["Rare"] = ITEM_QUALITY3_DESC
	L["Rare-Elite"] = ELITE .. "-" .. ITEM_QUALITY3_DESC
	L["Elite"] = ELITE
	L["Boss"] = BOSS
	-- short classifications
	L["Rare_short"] = "r"
	L["Rare-Elite_short"] = "r+"
	L["Elite_short"] = "+"
	L["Boss_short"] = "b"

	L["Feigned Death"] = "Feint la mort"
	L["Stealthed"] = "Camouflé"
	L["Soulstoned"] = "Âme conservée"

	L["Dead"] = DEAD
	L["Ghost"] = "Fantôme"
	L["Offline"] = PLAYER_OFFLINE
	L["Online"] = "En ligne"
	L["Combat"] = "Combat"
	L["Resting"] = "Repos"
	L["Tapped"] = "Touché"
	L["AFK"] = "ABS"
	L["DND"] = "NPD"

	L["True"] = "Vrai"

	L["Rage"] = RAGE
	L["Focus"] = FOCUS
	L["Energy"] = ENERGY
	L["Mana"] = MANA

	L["PvP"] = PVP
	L["FFA"] = "FFA"

	-- genders
	L["Male"] = MALE
	L["Female"] = FEMALE

	-- forms
	L["Bear"] = "Ours"
	L["Cat"] = "Félin"
	L["Moonkin"] = "Sélénien"
	L["Aquatic"] = "Aquatique"
	L["Flight"] = "Vol"
	L["Travel"] = "Voyage"
	L["Tree"] = "Arbre"

	L["Bear_short"] = "Ou"
	L["Cat_short"] = "Fé"
	L["Moonkin_short"] = "Sé"
	L["Aquatic_short"] = "Aq"
	L["Flight_short"] = "Vo"
	L["Travel_short"] = "Vy"
	L["Tree_short"] = "Ar"

	-- shortgenders
	L["Male_short"] = "h"
	L["Female_short"] = "f"

	L["Leader"] = "Chef"

	-- spell trees
	L["Hybrid"] = "Hybride" -- for all 3 trees
	L["Druid_Tree_1"] = "Equilibre"
	L["Druid_Tree_2"] = "Combat farouche"
	L["Druid_Tree_3"] = "Restauration"
	L["Hunter_Tree_1"] = "Maîtrise des bêtes"
	L["Hunter_Tree_2"] = "Précision"
	L["Hunter_Tree_3"] = "Survie"
	L["Mage_Tree_1"] = "Arcane"
	L["Mage_Tree_2"] = "Feu"
	L["Mage_Tree_3"] = "Givre"
	L["Paladin_Tree_1"] = "Sacré"
	L["Paladin_Tree_2"] = "Protection"
	L["Paladin_Tree_3"] = "Vindicte"
	L["Priest_Tree_1"] = "Discipline"
	L["Priest_Tree_2"] = "Sacré"
	L["Priest_Tree_3"] = "Ombre"
	L["Rogue_Tree_1"] = "Assassinat"
	L["Rogue_Tree_2"] = "Combat"
	L["Rogue_Tree_3"] = "Finesse"
	L["Shaman_Tree_1"] = "Elémentaire"
	L["Shaman_Tree_2"] = "Amélioration"
	L["Shaman_Tree_3"] = "Restauration"
	L["Warrior_Tree_1"] = "Armes"
	L["Warrior_Tree_2"] = "Fureur"
	L["Warrior_Tree_3"] = "Protection"
	L["Warlock_Tree_1"] = "Affliction"
	L["Warlock_Tree_2"] = "Démonologie"
	L["Warlock_Tree_3"] = "Destruction"
elseif GetLocale() == "zhTW" then
	-- races
	L["Blood Elf"] = "血精靈"
	L["Draenei"] = "德萊尼"
	L["Dwarf"] = "矮人"
	L["Gnome"] = "地精"
	L["Human"] = "人類"
	L["Night Elf"] = "夜精靈"
	L["Orc"] = "獸人"
	L["Tauren"] = "牛頭人"
	L["Troll"] = "食人妖"
	L["Undead"] = "不死族"

	-- short races
	L["Blood Elf_short"] = "血"
	L["Draenei_short"] = "萊"
	L["Dwarf_short"] = "矮"
	L["Gnome_short"] = "地"
	L["Human_short"] = "人"
	L["Night Elf_short"] = "夜"
	L["Orc_short"] = "獸"
	L["Tauren_short"] = "牛"
	L["Troll_short"] = "食"
	L["Undead_short"] = "不"

	-- classes
	L["Warrior"] = "戰士"
	L["Priest"] = "牧師"
	L["Mage"] = "法師"
	L["Shaman"] = "薩滿"
	L["Paladin"] = "聖騎士"
	L["Warlock"] = "術士"
	L["Druid"] = "德魯伊"
	L["Rogue"] = "盜賊"
	L["Hunter"] = "獵人"

	-- short classes
	L["Warrior_short"] = "戰"
	L["Priest_short"] = "牧"
	L["Mage_short"] = "法"
	L["Shaman_short"] = "薩"
	L["Paladin_short"] = "聖"
	L["Warlock_short"] = "術"
	L["Druid_short"] = "德"
	L["Rogue_short"] = "賊"
	L["Hunter_short"] = "獵"

	L["Player"] = PLAYER
	L["Target"] = TARGET
	L["Focus-target"] = FOCUSTARGET
	L["Mouse-over"] = "滑鼠目標"
	L["%s's pet"] = "%s的寵物"
	L["%s's target"] = "%s的目標"
	L["Party member #%d"] = "隊伍成員#%d"
	L["Raid member #%d"] = "團隊成員#%d"

	-- classifications
	L["Rare"] = "稀有"
	L["Rare-Elite"] = "稀有" .. "-" .. ELITE
	L["Elite"] = ELITE
	L["Boss"] = BOSS
	-- short classifications
	L["Rare_short"] = "稀"
	L["Rare-Elite_short"] = "稀+"
	L["Elite_short"] = "+"
	L["Boss_short"] = "首"

	L["Feigned Death"] = "假死"
	L["Stealthed"] = "潛行"
	L["Soulstoned"] = "靈魂被保存"

	L["Dead"] = DEAD
	L["Ghost"] = "鬼魂"
	L["Offline"] = PLAYER_OFFLINE
	L["Online"] = "上線"
	L["Combat"] = "戰鬥"
	L["Resting"] = "休息"
	L["Tapped"] = "選取"
	L["AFK"] = "暫離"
	L["DND"] = "勿擾"

	L["True"] = "真"

	L["Rage"] = RAGE
	L["Focus"] = FOCUS
	L["Energy"] = ENERGY
	L["Mana"] = MANA

	L["PvP"] = PVP
	L["FFA"] = PVPFFA

	-- genders
	L["Male"] = MALE
	L["Female"] = FEMALE

	-- forms
	L["Bear"] = "熊"
	L["Cat"] = "獵豹"
	L["Moonkin"] = "梟獸"
	L["Aquatic"] = "水棲"
	L["Flight"] = "飛行"
	L["Travel"] = "旅行"
	L["Tree"] = "樹"

	L["Bear_short"] = "熊"
	L["Cat_short"] = "豹"
	L["Moonkin_short"] = "梟"
	L["Aquatic_short"] = "水"
	L["Flight_short"] = "飛"
	L["Travel_short"] = "旅"
	L["Tree_short"] = "樹"

	-- shortgenders
	L["Male_short"] = "男"
	L["Female_short"] = "女"

	L["Leader"] = "隊長"

	-- spell trees
	L["Hybrid"] = "混合" -- for all 3 trees
	L["Druid_Tree_1"] = "平衡"
	L["Druid_Tree_2"] = "野性戰鬥"
	L["Druid_Tree_3"] = "恢復"
	L["Hunter_Tree_1"] = "野獸控制"
	L["Hunter_Tree_2"] = "射擊"
	L["Hunter_Tree_3"] = "生存"
	L["Mage_Tree_1"] = "秘法"
	L["Mage_Tree_2"] = "火焰"
	L["Mage_Tree_3"] = "冰霜"
	L["Paladin_Tree_1"] = "神聖"
	L["Paladin_Tree_2"] = "防護"
	L["Paladin_Tree_3"] = "懲戒"
	L["Priest_Tree_1"] = "戒律"
	L["Priest_Tree_2"] = "神聖"
	L["Priest_Tree_3"] = "暗影"
	L["Rogue_Tree_1"] = "刺殺"
	L["Rogue_Tree_2"] = "戰鬥"
	L["Rogue_Tree_3"] = "敏銳"
	L["Shaman_Tree_1"] = "元素"
	L["Shaman_Tree_2"] = "增強"
	L["Shaman_Tree_3"] = "恢復"
	L["Warrior_Tree_1"] = "武器"
	L["Warrior_Tree_2"] = "狂怒"
	L["Warrior_Tree_3"] = "防護"
	L["Warlock_Tree_1"] = "痛苦"
	L["Warlock_Tree_2"] = "惡魔學識"
	L["Warlock_Tree_3"] = "毀滅"

--***************************************
-- zhCN Chinese Simplify
-- 2007/09/19 CN3羽月 雪夜之狼
-- 请保留本翻译作者名 谢谢
-- E=mail:xionglingfeng@Gmail.com
-- Website:http://www.wowtigu.org  (Chs)
--***************************************
elseif GetLocale() == "zhCN" then
	-- races
	L["Blood Elf"] = "血精灵"
	L["Draenei"] = "德莱尼"
	L["Dwarf"] = "矮人"
	L["Gnome"] = "侏儒"
	L["Human"] = "人类"
	L["Night Elf"] = "暗夜精灵"
	L["Orc"] = "兽人"
	L["Tauren"] = "牛头人"
	L["Troll"] = "巨魔"
	L["Undead"] = "亡灵"

	-- short races
	L["Blood Elf_short"] = "血"
	L["Draenei_short"] = "德"
	L["Dwarf_short"] = "矮"
	L["Gnome_short"] = "侏"
	L["Human_short"] = "人"
	L["Night Elf_short"] = "暗"
	L["Orc_short"] = "兽"
	L["Tauren_short"] = "牛"
	L["Troll_short"] = "巨"
	L["Undead_short"] = "亡"

	-- classes
	L["Warrior"] = "战士"
	L["Priest"] = "牧师"
	L["Mage"] = "法师"
	L["Shaman"] = "萨满祭司"
	L["Paladin"] = "圣骑士"
	L["Warlock"] = "术士"
	L["Druid"] = "德鲁伊"
	L["Rogue"] = "潜行者"
	L["Hunter"] = "猎人"

	-- short classes
	L["Warrior_short"] = "战"
	L["Priest_short"] = "牧"
	L["Mage_short"] = "法"
	L["Shaman_short"] = "萨"
	L["Paladin_short"] = "圣"
	L["Warlock_short"] = "术"
	L["Druid_short"] = "德"
	L["Rogue_short"] = "贼"
	L["Hunter_short"] = "猎"

	L["Player"] = "玩家"
	L["Target"] = "目标"
	L["Focus-target"] = "焦点目标"
	L["Mouse-over"] = "鼠标目标"
	L["%s's pet"] = "%s的宠物"
	L["%s's target"] = "%s的目标"
	L["Party member #%d"] = "队伍成员#%d"
	L["Raid member #%d"] = "团队成员#%d"

	-- classifications
	L["Rare"] = "稀有"
	L["Rare-Elite"] = "稀有" .. "-" .. ELITE
	L["Elite"] = "精英"
	L["Boss"] = "Boss"
	-- short classifications
	L["Rare_short"] = "稀"
	L["Rare-Elite_short"] = "稀+"
	L["Elite_short"] = "+"
	L["Boss_short"] = "首"

	L["Feigned Death"] = "假死"
	L["Stealthed"] = "潜行"
	L["Soulstoned"] = "灵魂已保存"

	L["Dead"] = "死亡"
	L["Ghost"] = "鬼魂"
	L["Offline"] = "离线"
	L["Online"] = "上线"
	L["Combat"] = "战斗"
	L["Resting"] = "休息"
	L["Tapped"] = "选取"
	L["AFK"] = "暂离"
	L["DND"] = "勿扰"

	L["True"] = "真"

	L["Rage"] = RAGE
	L["Focus"] = "焦点"
	L["Energy"] = "能量"
	L["Mana"] = "魔法值"

	L["PvP"] = PVP
	L["FFA"] = PVPFFA

	-- genders
	L["Male"] = "男"
	L["Female"] = "女"

	-- forms
	L["Bear"] = "熊"
	L["Cat"] = "猎豹"
	L["Moonkin"] = "枭兽"
	L["Aquatic"] = "水栖"
	L["Flight"] = "飞行"
	L["Travel"] = "旅行"
	L["Tree"] = "树"

	L["Bear_short"] = "熊"
	L["Cat_short"] = "豹"
	L["Moonkin_short"] = "枭"
	L["Aquatic_short"] = "水"
	L["Flight_short"] = "飞"
	L["Travel_short"] = "旅"
	L["Tree_short"] = "树"

	-- shortgenders
	L["Male_short"] = "男"
	L["Female_short"] = "女"

	L["Leader"] = "队长"

	-- spell trees
	L["Hybrid"] = "Hybrid" -- for all 3 trees
	L["Druid_Tree_1"] = "平衡"
	L["Druid_Tree_2"] = "野性战斗"
	L["Druid_Tree_3"] = "恢复"
	L["Hunter_Tree_1"] = "野兽掌握"
	L["Hunter_Tree_2"] = "射击"
	L["Hunter_Tree_3"] = "生存"
	L["Mage_Tree_1"] = "奥术"
	L["Mage_Tree_2"] = "火焰"
	L["Mage_Tree_3"] = "冰霜"
	L["Paladin_Tree_1"] = "神圣"
	L["Paladin_Tree_2"] = "防护"
	L["Paladin_Tree_3"] = "惩戒"
	L["Priest_Tree_1"] = "戒律"
	L["Priest_Tree_2"] = "神圣"
	L["Priest_Tree_3"] = "暗影"
	L["Rogue_Tree_1"] = "刺杀"
	L["Rogue_Tree_2"] = "战斗"
	L["Rogue_Tree_3"] = "敏锐"
	L["Shaman_Tree_1"] = "元素战斗"
	L["Shaman_Tree_2"] = "增强"
	L["Shaman_Tree_3"] = "恢复"
	L["Warrior_Tree_1"] = "武器"
	L["Warrior_Tree_2"] = "狂怒"
	L["Warrior_Tree_3"] = "防御"
	L["Warlock_Tree_1"] = "痛苦"
	L["Warlock_Tree_2"] = "恶魔学识"
	L["Warlock_Tree_3"] = "毁灭"
end

setmetatable(L, {__index = function(self, key)
--	local _, ret = pcall(error, ("Error indexing L[%q]"):format(tostring(key)), 2)
--	geterrorhandler()(ret)
	self[key] = key
	return key
end})
DogTag.L = L

local AceLibrary = AceLibrary
local BabbleSpell = AceLibrary and AceLibrary:HasInstance("Babble-Spell-2.2") and AceLibrary("Babble-Spell-2.2")
local Threat

local _G = _G
local RAID_CLASS_COLORS = RAID_CLASS_COLORS
local UNKNOWN = UNKNOWN or "Unknown"
local unpack = unpack
local pcall = pcall
local select = select
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local UnitIsFriend = UnitIsFriend
local UnitIsAFK = UnitIsAFK
local UnitIsConnected = UnitIsConnected
local UnitIsDeadOrGhost = UnitIsDeadOrGhost
local UnitInRaid = UnitInRaid
local UnitInParty = UnitInParty
local UnitExists = UnitExists
local UnitIsVisible = UnitIsVisible
local UnitIsPlayer = UnitIsPlayer
local UnitPlayerControlled = UnitPlayerControlled
local time = time
local UnitName = UnitName
local rawget = rawget
local pairs = pairs
local ipairs = ipairs
local next = next
local setmetatable = setmetatable
local type = type
local assert = assert
local tonumber = tonumber
local tostring = tostring
local string_gsub = string.gsub
local math_floor = math.floor
local table_sort = table.sort
local table_concat = table.concat
local error = error
local loadstring = loadstring
local GetTime = GetTime
local GetNumGuildMembers = GetNumGuildMembers
local GetGuildRosterInfo = GetGuildRosterInfo

local function GetNameServer(name, server)
	if name then
		if server and server ~= "" then
			return name .. "-" .. server
		else
			return name
		end
	else
		return nil
	end
end

local newList, newDict, newSet, del
do
	local pool = {}
	function newList(...)
		local t = next(pool)
		if t then
			pool[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
		else
			t = { ... }
		end
		return t
	end
	function newDict(...)
		local t = next(pool)
		if t then
			pool[t] = nil
		else
			t = {}
		end
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function newSet(...)
		local t = next(pool)
		if t then
			pool[t] = nil
		else
			t = {}
		end
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	function del(t)
		if not t then
			error("Bad argument #1 to `del'. Expected table, got nil.", 2)
		end
		pool[t] = true
		for k in pairs(t) do
			t[k] = nil
		end
		t[''] = true
		t[''] = nil
		return nil
	end
end
DogTag.newList, DogTag.newDict, DogTag.newSet, DogTag.del = newList, newDict, newSet, del

local precondition, postcondition, argCheck
if Rock then
	precondition, postcondition, argCheck = Rock:GetContractFunctions(MAJOR_VERSION, "precondition", "postcondition", "argCheck")
else
	precondition, postcondition, argCheck = function() end, function() end, function() end
end

local colors = {}
for class, data in pairs(RAID_CLASS_COLORS) do
	colors[class] = { data.r, data.g, data.b, }
end
colors.unknown = { 0.8, 0.8, 0.8 }
colors.hostile = { 226/255, 45/255, 75/255 }
colors.neutral = { 1, 1, 34/255 }
colors.friendly = { 0.2, 0.8, 0.15 }
colors.civilian = { 48/255, 113/255, 191/255 }
colors.tapped = { 0.5, 0.5, 0.5 }
colors.dead = { 0.6, 0.6, 0.6 }
colors.disconnected = { 0.7, 0.7, 0.7 }
colors.inCombat = { 1, 0, 0 }
colors.petHappy = { 0, 1, 0 }
colors.petNeutral = { 1, 1, 0 }
colors.petAngry = { 1, 0, 0 }
colors.rage = { 226/255, 45/255, 75/255 }
colors.energy = { 1, 220/255, 25/255 }
colors.focus = { 1, 210/255, 0 }
colors.mana = { 48/255, 113/255, 191/255 }
colors.minHP = { 1, 0, 0 }
colors.midHP = { 1, 1, 0 }
colors.maxHP = { 0, 1, 0 }

local lastMouseFocus

local IsLegitimateUnit = {player=true,pet=true,target=true,focus=true,mouseover=true}
DogTag.IsLegitimateUnit = IsLegitimateUnit
for i = 1, 4 do
	IsLegitimateUnit['party' .. i] = true
	IsLegitimateUnit['partypet' .. i] = true
end
for i = 1, 40 do
	IsLegitimateUnit['raid' .. i] = true
	IsLegitimateUnit['raidpet' .. i] = true
end
local IsNormalUnit = {}
for k in pairs(IsLegitimateUnit) do
	IsNormalUnit[k] = true
end
setmetatable(IsLegitimateUnit, {__index=function(self, unit)
	if type(unit) ~= "string" then
		return false
	end
	if not unit:find("target$") then
		self[unit] = false
		return false
	end
	local nonTarget = unit:sub(1, -7)
	local good = self[nonTarget]
	self[unit] = good
	return good
end})

local currentAuras, currentDebuffTypes, currentNumDebuffs
local function func(self, unit)
	local t = newList()
	local u = newList()
	for i = 1, 40 do
		local name = UnitBuff(unit, i)
		if not name then
			break
		end
		t[name] = (t[name] or 0) + 1
	end
	local numDebuffs = 0
	local isFriend = UnitIsFriend("player", unit)
	for i = 1, 40 do
		local name, _, _, _, dispelType = UnitDebuff(unit, i)
		if not name then
			break
		end
		numDebuffs = numDebuffs + 1
		t[name] = (t[name] or 0) + 1
		if isFriend and dispelType then
			u[dispelType] = true
		end
	end
	currentAuras[unit] = t
	currentDebuffTypes[unit] = u
	currentNumDebuffs[unit] = numDebuffs
	return self[unit]
end
local x = {__index=func}
currentAuras = setmetatable({},x)
currentDebuffTypes = setmetatable({},x)
currentNumDebuffs = setmetatable({},x)
x = nil

local guildNotes, officerNotes
local function refreshGuildNotes(auto)
	if auto and getmetatable(guildNotes) then
		return
	end
	for k in pairs(guildNotes) do
		guildNotes[k] = nil
	end
	for k in pairs(officerNotes) do
		officerNotes[k] = nil
	end
	if not IsInGuild() then
		return nil
	end
	for i = 1, GetNumGuildMembers(true) do
		local name, _, _, _, _, _, note, officernote = GetGuildRosterInfo(i)
		if note then
			note = note:trim()
			if note == "" then
				note = nil
			end
		end
		if officernote then
			officernote = officernote:trim()
			if officernote == "" then
				officernote = nil
			end
		end
		guildNotes[name] = note
		officerNotes[name] = officernote
	end
end
local x = {__index=function(self, name)
	refreshGuildNotes()
	setmetatable(guildNotes, nil)
	setmetatable(officerNotes, nil)
	return self[name]
end}
guildNotes = setmetatable({},x)
officerNotes = setmetatable({},x)

local castData = setmetatable({}, {__index=function(self, unit)
	local data = newList()
	self[unit] = data
	
	local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo(unit)
	local casting = true
	
	if not spell then
		spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo(unit)
		casting = false
		
		if not spell then
			return data
		end
	end
	data.spell = spell
	data.rank = rank and tonumber(rank:match("%d+"))
	data.displayName = displayName
	data.icon = icon
	data.startTime = startTime / 1000
	data.endTime = endTime / 1000
	data.delay = 0
	data.casting = casting
	data.stopTime = nil
	data.stopMessage = nil
	
	return data
end})

local tt = CreateFrame("GameTooltip")
local UIParent = UIParent
tt:SetOwner(UIParent, "ANCHOR_NONE")
tt.left = {}
tt.right = {}
for i=1,30 do
	tt.left[i] = tt:CreateFontString()
	tt.left[i]:SetFontObject(GameFontNormal)
	tt.right[i] = tt:CreateFontString()
	tt.right[i]:SetFontObject(GameFontNormal)
	tt:AddFontStrings(tt.left[i], tt.right[i])
end
local nextTime = 0
local lastName
local lastUnit
local function updateTT(unit)
	local name = UnitName(unit)
	local time = GetTime()
	if lastUnit == unit and lastName == name and nextTime < time then
		return
	end
	lastUnit = unit
	lastName = name
	nextTime = time + 1
	tt:ClearLines()
	tt:SetUnit(unit)
	if not tt:IsOwned(UIParent) then
		tt:SetOwner(UIParent, "ANCHOR_NONE")
	end
end

local LEVEL_start = "^" .. (type(LEVEL) == "string" and LEVEL or "Level")
local function FigureNPCGuild(unit)
	updateTT(unit)
	local left_2 = tt.left[2]:GetText()
	if not left_2 or left_2:find(LEVEL_start) then
		return nil
	end
	return left_2
end

local factionList = {}

local PVP = type(PVP) == "string" and PVP or "PvP"
local function FigureFaction(unit)
	local _, faction = UnitFactionGroup(unit)
	if UnitPlayerControlled(unit) or UnitIsPlayer(unit) then
		return faction
	end

	updateTT(unit)
	local left_2 = tt.left[2]:GetText()
	local left_3 = tt.left[3]:GetText()
	if not left_2 or not left_3 then
		return faction
	end
	local hasGuild = not left_2:find(LEVEL_start)
	local factionText = not hasGuild and left_3 or tt.left[4]:GetText()
	if factionText == PVP then
		return faction
	end
	if factionList[factionText] or faction then
		return factionText
	end
end

local function FigureZone(unit)
	if UnitIsVisible(unit) then
		return nil
	end
	if not UnitIsConnected(unit) then
		return nil
	end
	updateTT(unit)
	local left_2 = tt.left[2]:GetText()
	local left_3 = tt.left[3]:GetText()
	if not left_2 or not left_3 then
		return nil
	end
	local hasGuild = not left_2:find(LEVEL_start)
	local factionText = not hasGuild and left_3 or tt.left[4]:GetText()
	if factionText == PVP then
		factionText = nil
	end
	local hasFaction = factionText and not UnitPlayerControlled(unit) and not UnitIsPlayer(unit) and (UnitFactionGroup(unit) or factionList[factionText])
	if hasGuild and hasFaction then
		return tt.left[5]:GetText()
	elseif hasGuild or hasFaction then
		return tt.left[4]:GetText()
	else
		return left_3
	end
end

local RockComm

local talentReceptor, talentSpecs, talentSpecNames
local addonVersionReceptor

local triedTalentWhisper = {} -- don't whisper twice
local lastInspectTime = 0 -- based on GetTime()
local lastInspectName = nil
local lastInspectClass = nil
local inspectionQueue = {}

local function QueryTalents(name)
	-- assume talentSpecs[".current"] is a proper unit and UnitIsPlayer(unit)
	
	if not RockComm then
		RockComm = Rock and Rock("LibRockComm-1.0", false, true)
		if RockComm then
			RockComm:AddTalentReceptor(talentReceptor)
			RockComm:AddAddonVersionReceptor(addonVersionReceptor)
		end
	end
	
	local unit = talentSpecs[".current"]
	
	if RockComm and not triedTalentWhisper[name] then
		triedTalentWhisper[name] = true
		if UnitIsFriend("player", unit) then
			RockComm:QueryTalents("WHISPER", name)
		end
	end
	
	inspectionQueue[name] = true
	if (not InspectFrame or not InspectFrame.unit) and UnitIsVisible(unit) and UnitIsConnected(unit) and (GetTime() - lastInspectTime > 15 or not lastInspectName) then
		lastInspectName = name
		local _
		_, lastInspectClass = UnitClass(unit)
		lastInspectTime = GetTime()
		NotifyInspect(unit)
	end
end

local mt = {__index = function(self, key)
	QueryTalents(key)
	return rawget(self, key) or false
end}
talentSpecs = setmetatable({}, mt)
talentSpecNames = setmetatable({}, mt)

local function better_UnitLevel(unit)
	local level = UnitLevel(unit)
	if level > 0 or not UnitIsPlayer(unit) then
		return level
	end
	local name, server = UnitName(unit)
	if server and server ~= "" then
		name = name .. "-" .. server
	end
	talentSpecs[".current"] = unit
	local value = talentSpecs[name]
	if not value then
		return 0
	end
	local a, b, c = value:match("^(%d+)/(%d+)/(%d+)$")
	if not a then
		return 0
	end
	local total = a + b + c
	return total + 9
end

local mt = {__index = function(self, key)
	self[key] = false
	if not RockComm then
		RockComm = Rock and Rock("LibRockComm-1.0", false, true)
		if not RockComm then
			return self[key]
		end
		RockComm:AddTalentReceptor(talentReceptor)
		RockComm:AddAddonVersionReceptor(addonVersionReceptor)
	end
	RockComm:QueryAddonVersion(self[0], "WHISPER", key)
	return self[key]
end}

local addonVersions = setmetatable({}, {__index = function(self, key)
	local t = setmetatable({ [0] = key }, mt)
	self[key] = t
	return t
end})

local Modifiers; Modifiers = {}
-- when writing modifiers, inputs can be a number or a string. Also, any of these can be returned, as well as nil, which means a blank string.
-- the modifier itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- note: can take an argument, either nil, a number, a string.


local Tags = {}
-- When writing tags, `unit` is modified before the call, that contains the unit id.
-- Also, nil (which translates to empty string), a number, or a string can be returned.
-- the tag itself can be in the form of a function or a table containing a function and an event string (or a table of event strings). The events specified will cause the string to update upon that event.
-- Note: can take an argument, either nil, a number, or a string.

Tags._ = {[[value = "_ tag referenced improperly"]], arg = "string;number;nil", ret = "string;number;nil"}

local function sortStringList(s)
	if not s then
		return nil
	end
	local list = newList((";"):split(s))
	table_sort(list)
	local q = table_concat(list, ';')
	list = del(list)
	return q
end

local TagEvents = {}
local ModifierEvents = {}
local TagAliases = {}
local ModifierAliases = {}
local TagAliasFunctions = {}
local ModifierAliasFunctions = {}
local TagArguments = {}
local ModifierArguments = {}
local TagReturns = {}
local ModifierReturns = {}
local ModifierValues = {}
local TagGlobals = {}
local ModifierGlobals = {}
local TagDocuments = {}
local ModifierDocuments = {}
local TagDocumentCategories = {}
local ModifierDocumentCategories = {}
local TagDocumentExamples = {}
local ModifierDocumentExamples = {}

local correctCasing = setmetatable({},{__mode='kv',__index=function(self, tag)
	if not tag then
		return nil
	end
	if Tags[tag] or Modifiers[tag] or TagAliases[tag] or ModifierAliases[tag] then
		self[tag] = tag
		return tag
	end
	local tag_lower = tag:lower()
	for k in pairs(Tags) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	for k in pairs(Modifiers) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	for k in pairs(TagAliases) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	for k in pairs(ModifierAliases) do
		if k:lower() == tag_lower then
			self[tag] = k
			return k
		end
	end
	self[tag] = tag
	return tag
end})

local function toliteral(value)
	if type(value) == "number" then
		return ("(%s)"):format(value)
	else
		return ("(%q)"):format(value)
	end
end

local function unpackAndDel(t, start, finish)
	if not start then
		start = 1
	elseif not finish then
		finish = #t
	end
	if start > finish then
		t = del(t)
		return
	end
	return t[start], unpackAndDel(t, start + 1, finish)
end

local ops = {
	{ ["^"] = "Pow" },
	{ ["*"] = "Mult", ["/"] = "Div", ["%"] = "Mod" },
	{ ["+"] = "Add", ["-"] = "Sub" },
	{ ["<"] = "IsLess", [">"] = "IsGreater", ["<="] = "IsLessEqual", [">="] = "IsGreaterEqual", ["="] = "IsEqual", ["~="] = "IsNotEqual" }
}
local opSet = { ["!"] = true, ["?"] = true }
for _, opList in ipairs(ops) do
	for k in pairs(opList) do
		opSet[k] = true
	end
end

local function splitCode(code, inner)
	code = code:trim()
	if code:sub(1, 1) == "[" then
		local alpha, bravo, charlie = code:match("^(%b[])%s*([<>]=)%s*(.*)$")
		if not alpha then
			alpha, bravo, charlie = code:match("^(%b[])%s*(~=)%s*(.*)$")
		end
		if not alpha then
			alpha, bravo, charlie = code:match("^(%b[])%s*([:!%?%+%-%*/^%%<>=])%s*(.*)$")
		end
		if alpha then
			return "_(" .. alpha:trim() .. ")", bravo:trim(), splitCode(charlie, true)
		end
		local alpha = code:match("^(%b[])$")
		if alpha then
			if inner then
				return "_(" .. alpha:trim() .. ")"
			else
				return splitCode(alpha:sub(2, -2):trim())
			end
		end
	end
	local a, _, nextSymbol = code:find("^.-([:!%?%+%-%*/^%%<>=%(])")
	local b, _, nextSymbol_ = code:find("^.-(~=)")
	if b and (not a or b < a) then
		nextSymbol = nextSymbol_
	end
	if nextSymbol == "(" then
		local alpha, bravo, charlie = code:match("^([^%(]*%b())%s*([<>]=)%s*(.*)$")
		if not alpha then
			alpha, bravo, charlie = code:match("^([^%(]*%b())%s*(~=)%s*(.*)$")
		end
		if not alpha then
			alpha, bravo, charlie = code:match("^([^%(]*%b())%s*([:!%?%+%-%*/^%%<>=])%s*(.*)$")
		end
		if alpha then
			return alpha:trim(), bravo:trim(), splitCode(charlie, true)
		end
		local alpha = code:match("^([^%(]*%b())$")
		if alpha then
			return alpha:trim()
		end
	end
	local alpha, bravo, charlie = code:match("^([^:!%?%+%-%*/^%%<>=]*)%s*([<>]=)%s*(.*)$")
	if not alpha then
		alpha, bravo, charlie = code:match("^([^:!%?%+%-%*/^%%<>=]*)%s*(~=)%s*(.*)$")
	end
	if not alpha then
		alpha, bravo, charlie = code:match("^([^:!%?%+%-%*/^%%<>=]*)%s*([:!%?%+%-%*/^%%<>=])%s*(.*)$")
	end
	if not alpha then
		return code
	end
	if charlie:sub(1, 1) == "[" then
		for _, oplist in ipairs(ops) do
			if oplist[bravo] then
				return alpha:trim(), bravo:trim(), charlie:trim()
			end
		end
	end
	return alpha:trim(), bravo:trim(), splitCode(charlie, true)
end
_G.splitCode = splitCode

local FixCasing

local function _splitParam(code)
	code = code:trim()
	local negate, tag, param = code:match("^(~*)%s*([a-zA-Z_#0-9%s]+)%s*(%b())$")
	if param then
		param = param:sub(2, -2)
	end
	if tag then
		local t, unit = tag:match("^([a-zA-Z_]+)%s*#%s*([a-zA-Z0-9_]+)$")
		if t then
			return t:trim(), param, ((negate:len() % 2) == 1 and true or false), unit:trim():lower()
		else
			return tag:trim(), param, ((negate:len() % 2) == 1 and true or false), nil
		end
	else
		negate, tag = code:match("^(~*)%s*(.+)$")
		if tag then
			local t, unit = tag:match("^([a-zA-Z_]+)%s*#%s*([a-zA-Z0-9_]+)$")
			if t then
				return t:trim(), nil, ((negate:len() % 2) == 1 and true or false), unit:trim():lower()
			else
				return tag:trim(), nil, ((negate:len() % 2) == 1 and true or false), nil
			end
		else
			return code:trim(), nil, nil, nil
		end
	end
end
local function splitParam(code)
	local code, param, negate, unit = _splitParam(code)
	code = correctCasing[code]
	if type(param) == "string" and param:find("%[") and param:find("%]") then
		param = FixCasing(param)
	end
	if negate then
		local neg_code = '~' .. code
		if Tags[neg_code] or Modifiers[neg_code] or TagAliases[neg_code] or ModifierAliases[neg_code] then
			return neg_code, param, false, unit
		end
	end
	return code, param, negate, unit
end

local function joinParam(tag, param, negate, unit)
	if tag == "_" and param and param:find("^%[.*%]$") then
		local inner = param:sub(2, -2)
		if not inner:find("[%[%]:!%?%+%-%*/^%%<>=]") and not inner:find("~=") then
			return inner
		end
		return param
	end
	local t = newList()
	if negate then
		t[#t+1] = "~"
	end
	t[#t+1] = tag
	if unit then
		t[#t+1] = "#"
		t[#t+1] = unit
	end
	if param then
		t[#t+1] = "("
		t[#t+1] = param
		t[#t+1] = ")"
	end
	local s = table_concat(t)
	t = del(t)
	return s
end

local parse, compileTree
local function _compileTree(code, globals, cachedTags, doneCachedTags, lastPossibleReturns)
	if type(code) == "string" then
		return code, 0, lastPossibleReturns
	end
	local ends
	local tmp
	code[1], ends, tmp = _compileTree(code[1], globals, cachedTags, doneCachedTags, lastPossibleReturns)
	if not code[1] then
		code = del(code)
		lastPossibleReturns = del(lastPossibleReturns)
		return nil, ends
	end
	lastPossibleReturns = tmp
	local t = newList()
	local finalPossibleReturns = newList()
	if not code[2] then
		local tag, param, negate, unit = splitParam(code[1])

		local func = Tags[tag]
		if func then
			local hasCache = false
			local firstCache = not doneCachedTags[code[1]]
			if cachedTags[code[1]] then
				hasCache = cachedTags[code[1]]
				doneCachedTags[code[1]] = true
				if not firstCache then
					t[#t+1] = "if "
					t[#t+1] = hasCache
					t[#t+1] = " ~= NIL then value = "
					t[#t+1] = hasCache
					t[#t+1] = ";"
					t[#t+1] = " else "
				end
			elseif code[1]:find("^~") and cachedTags[code[1]:sub(2)] then
				hasCache = cachedTags[code[1]:sub(2)]
				if tag:find("^~") then
					tag = tag:sub(2)
					negate = not negate
					func = Tags[tag]
					assert(func)
				end
				doneCachedTags[code[1]:sub(2)] = true
				if not firstCache then
					t[#t+1] = "if "
					t[#t+1] = hasCache
					t[#t+1] = " ~= NIL then value = not "
					t[#t+1] = hasCache
					t[#t+1] = (" and %q;"):format(L["True"])
					t[#t+1] = " else "
				end
			end
			if param then
				if not TagArguments[tag] then
					t = del(t)
					code = del(code)
					lastPossibleReturns = del(lastPossibleReturns)
					finalPossibleReturns = del(finalPossibleReturns)
					return nil, ("%q Unexpected parameter to %q"):format(param, tag)
				end
				if TagArguments[tag]:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not TagArguments[tag]:find("string")  and (not param:find("%[") or not param:find("%]")) then
						-- must be number
						t = del(t)
						code = del(code)
						lastPossibleReturns = del(lastPossibleReturns)
						finalPossibleReturns = del(finalPossibleReturns)
						return nil, ("%q Bad param to %q. Expected number."):format(param, tag)
					end
				end
			else
				if TagArguments[tag] and not TagArguments[tag]:find("nil") then
					t = del(t)
					code = del(code)
					lastPossibleReturns = del(lastPossibleReturns)
					finalPossibleReturns = del(finalPossibleReturns)
					return nil, ("%q Expected parameter"):format(tag)
				end
			end
			if unit and unit ~= "player" then
				if not IsLegitimateUnit[unit] then
					t = del(t)
					code = del(code)
					lastPossibleReturns = del(lastPossibleReturns)
					finalPossibleReturns = del(finalPossibleReturns)
					return nil, ("%q Unknown unit"):format(unit)
				end
				t[#t+1] = ("if UnitExists(%q) then "):format(unit)
				ends = ends + 1
			end
			if negate and not TagReturns[tag]:find("nil") then
				if hasCache then
					t[#t+1] = " end;"
				end
				t[#t+1] = "value = nil;"
				finalPossibleReturns["nil"] = true
			else
				local g = TagGlobals[tag]
				if g then
					g = newList((";"):split(g))
					for _,v in ipairs(g) do
						globals[v] = true
					end
					g = del(g)
				end

				if type(func) == "function" then
					local data = newList()
					data.isMod = false
					data.unit = unit
					data.unit_str = unit and ("(%q)"):format(unit) or "(unit)"
					data.arg = param
					data.arg_str = param and toliteral(param) or "(nil)"
					data.arg_string = param and ("(%q)"):format(param) or '("nil")'
					data.argtype = ("(%q)"):format(type(param))
					local hash = newDict('func', func(data))
					func = hash.func
					if hash.globals then
						g = newList((";"):split(hash.globals))
						for _,v in ipairs(g) do
							globals[v] = true
						end
						g = del(g)
					end
					hash = del(hash)
					data = del(data)
				end
				if param then
					if type(param) == "string" and param:find("%[") and param:find("%]") then
						local single = param:find("^%[[^%[%]]+%]$") and tag ~= "_"
						local possibleReturns

						if not single then
							if tag ~= "_" then
								t[#t+1] = "local v = value;"
							end

							t[#t+1] = "" -- placeholder

							local t_id = #t
							local madeT = false
							local t_numstart = math.random(1, 10000000)
							local t_num = t_numstart
							local t_literal = newList()
							local st = 1
							t[#t+1] = [[do ]]
							while st <= param:len() do
								local start, fin, left, literal = param:find("^(.-)(%b[])", st)
								if not start then
									local rest = param:sub(st)
									if rest:len() > 0 then
										if not madeT then
											t[#t+1] = ("value = %q"):format(rest)
											if possibleReturns then
												possibleReturns = del(possibleReturns)
											end
											possibleReturns = newSet("string")
										else
											t_literal[t_num+1] = rest
										end
									end
									break
								end
								if not madeT and (left:len() > 0 or start ~= 1 or fin ~= param:len()) then
									madeT = true
								end
								local inner = literal:sub(2, -2)
								st = fin+1
								if left:len() > 0 then
									t_literal[t_num] = left
								end

								local u = newList()
								if madeT then
									u[#u+1] = "do "
								end
								u[#u+1] = "value = nil;"

								local result, ret = compileTree(parse(splitCode(inner)), globals, cachedTags, doneCachedTags)
								u[#u+1] = result

								if madeT then
									t_num = t_num + 1
									u[#u+1] = ([[t_%d = value or '';]]):format(t_num)
									u[#u+1] = [[end;]]
									if ret then
										ret = del(ret)
									end
								else
									possibleReturns = ret
								end
								local s = table_concat(u)
								u = del(u)
								t[#t+1] = s
							end
							if madeT then
								local u = newList()
								for i = t_numstart+1, t_num do
									u[i] = ("local t_%d = '';"):format(i)
								end
								t[t_id] = table_concat(u)
								u = del(u)
								t[#t+1] = [[value = ]]
								if t_literal[0] then
									t[#t+1] = ("%q"):format(t_literal[0])
									t[#t+1] = " .. "
								end
								for i = t_numstart+1, t_num+1 do
									if i <= t_num then
										t[#t+1] = "t_"
										t[#t+1] = i
										t[#t+1] = " .. "
									end
									if t_literal[i] then
										t[#t+1] = ("%q"):format(t_literal[i])
										t[#t+1] = " .. "
									end
								end
								t[#t] = [[;]]

								possibleReturns = newSet("string")
							end
							t_literal = del(t_literal)
							t[#t+1] = [[end;]]
						end
						if tag ~= "_" then
							if single then
								local result, ret = compileTree(parse(splitCode(param:sub(2, -2))), globals, cachedTags, doneCachedTags)
								result = result:gsub("%f[A-Za-z0-9_]value%f[^A-Za-z0-9_]", "arg")
								possibleReturns = ret
								t[#t+1] = "local arg;"
								t[#t+1] = result
							else
								t[#t+1] = "local arg = value;"
							end
							local set = newSet((";"):split(TagArguments[tag] or 'nil'))
							if set["nil"] then
								if set.number then
									if set.string then
										-- do nothing
									else
										if (not possibleReturns["nil"] and not possibleReturns.number) or possibleReturns.string then
											t[#t+1] = "arg = tonumber(arg);"
										end
									end
								else
									if set.string then
										if possibleReturns.number then
											if possibleReturns["nil"] then
												t[#t+1] = "if arg then arg = tostring(arg) end;"
											else
												t[#t+1] = "arg = tostring(arg);"
											end
										end
									else
										error(("Bad argument list. Can't just return nil. %q"):format(tag))
									end
								end
							else
								if set.number then
									if set.string then
										if possibleReturns["nil"] then
											t[#t+1] = "if not arg then arg = '' end;"
										end
									else
										if possibleReturns["nil"] or possibleReturns.string then
											t[#t+1] = "arg = tonumber(arg) or 0;"
										end
									end
								else
									if set.string then
										if not possibleReturns["nil"] then
											if possibleReturns.number then
												t[#t+1] = "arg = tostring(arg);"
											end
										else
											if possibleReturns.number then
												t[#t+1] = "if arg then arg = tostring(arg) else arg = '' end;"
											else
												t[#t+1] = "if not arg then arg = '' end;"
											end
										end
									else
										error(("Bad argument list. Can't just return nil. %q"):format(tag))
									end
								end
							end
							set = del(set)
							t[#t+1] = "value = v;"
						end
						if possibleReturns then
							possibleReturns = del(possibleReturns)
						end
						func = func:gsub("${arg}", "(arg)")
						func = func:gsub("%(${arg}%)", "(arg)")
						local pos = func:find("${arg:string}")
						if pos and func:find("${arg:string}", pos+1) then
							t[#t+1] = "local arg_string = tostring(arg);"
							func = func:gsub("%(${arg:string}%)", "(arg_string)")
							func = func:gsub("${arg:string}", "(arg_string)")
						else
							func = func:gsub("%(${arg:string}%)", "(tostring(arg))")
							func = func:gsub("${arg:string}", "(tostring(arg))")
						end
						local pos = func:find("${argtype}")
						if pos and func:find("${argtype}", pos+1) then
							t[#t+1] = "local argtype = type(arg);"
							func = func:gsub("%(${argtype}%)", "(argtype)")
							func = func:gsub("${argtype}", "(argtype)")
						else
							func = func:gsub("%(${argtype}%)", "(type(arg))")
							func = func:gsub("${argtype}", "(type(arg))")
						end
					else
						func = func:gsub("%(${arg}%)", toliteral(param))
						func = func:gsub("${arg}", toliteral(param))
						func = func:gsub("%(${arg:string}%)", ("(%q)"):format(param))
						func = func:gsub("${arg:string}", ("(%q)"):format(param))
						func = func:gsub("%(${argtype}%)", ("(%q)"):format(type(param)))
						func = func:gsub("${argtype}", ("(%q)"):format(type(param)))
					end
				else
					func = func:gsub("%(${arg}%)", "(nil)")
					func = func:gsub("${arg}", "(nil)")
					func = func:gsub("%(${arg:string}%)", '("nil")')
					func = func:gsub("${arg:string}", '("nil")')
					func = func:gsub("%(${argtype}%)", '("nil")')
					func = func:gsub("${argtype}", '("nil")')
				end
				if unit then
					func = func:gsub("%(${unit}%)", ("(%q)"):format(unit))
					func = func:gsub("${unit}", ("(%q)"):format(unit))
				else
					func = func:gsub("%(${unit}%)", "(unit)")
					func = func:gsub("${unit}", "(unit)")
				end
				if tag ~= "_" then
					t[#t+1] = func
					if func ~= "" then
						t[#t+1] = ";"
					end
				end
				if hasCache then
					t[#t+1] = hasCache
					t[#t+1] = " = value;"
				end
				if negate then
					t[#t+1] = ("value = not value and %q;"):format(L["True"])
					finalPossibleReturns["string"] = true
					finalPossibleReturns["nil"] = true
				else
					local tagRet = newList((';'):split(TagReturns[tag]))
					for _,v in ipairs(tagRet) do
						finalPossibleReturns[v] = true
					end
					tagRet = del(tagRet)
				end
				if hasCache and not firstCache then
					t[#t+1] = " end;"
				end
			end
		else
			t = del(t)
			code = del(code)
			lastPossibleReturns = del(lastPossibleReturns)
			finalPossibleReturns = del(finalPossibleReturns)
			return nil, ("%q Unknown tag"):format(tag)
		end
	elseif code[2] == "?" then
		t[#t+1] = code[1]
		t[#t+1] = ("if value then value = nil;")
		local q, e = _compileTree(code[3], globals, cachedTags, doneCachedTags, newSet("nil"))
		if not q then
			t = del(t)
			code = del(code)
			lastPossibleReturns = del(lastPossibleReturns)
			finalPossibleReturns = del(finalPossibleReturns)
			return nil, e
		end
		t[#t+1] = q
		for i = 1, e do
			t[#t+1] = "end;"
		end
		if code[4] == "!" then
			t[#t+1] = ("else ")
			local q, e = _compileTree(code[5], globals, cachedTags, doneCachedTags, newSet("nil"))
			if not q then
				t = del(t)
				code = del(code)
				lastPossibleReturns = del(lastPossibleReturns)
				finalPossibleReturns = del(finalPossibleReturns)
				return nil, e
			end
			t[#t+1] = q
			for i = 1, e do
				t[#t+1] = "end;"
			end
		end
		t[#t+1] = ("end;")
	elseif code[2] == ":" then
		t[#t+1] = code[1]
		local mod, param, negate, unit = splitParam(code[3])
		local func = Modifiers[mod] or Tags[mod]
		if func then
			local isMod = func == Modifiers[mod]
			local hasCache = false
			local firstCache = not doneCachedTags[code[3]]
			if param then
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if not arg then
					t = del(t)
					code = del(code)
					lastPossibleReturns = del(lastPossibleReturns)
					finalPossibleReturns = del(finalPossibleReturns)
					return nil, ("%q Unexpected param to %q"):format(param, mod)
				end

				if arg:find("number") then
					param = tonumber(param) or param
					if type(param) ~= "number" and not arg:find("string") and (not param:find("%[") or not param:find("%]")) then
						-- must be number
						t = del(t)
						code = del(code)
						lastPossibleReturns = del(lastPossibleReturns)
						finalPossibleReturns = del(finalPossibleReturns)
						return nil, ("%q Bad param to %q. Expected number."):format(param, mod)
					end
				end
			else
				local arg = (isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod])
				if arg and not arg:find("nil") then
					t = del(t)
					code = del(code)
					lastPossibleReturns = del(lastPossibleReturns)
					finalPossibleReturns = del(finalPossibleReturns)
					return nil, ("%q Expected parameter"):format(mod)
				end
			end
			if unit and not IsLegitimateUnit[unit] then
				t = del(t)
				code = del(code)
				lastPossibleReturns = del(lastPossibleReturns)
				finalPossibleReturns = del(finalPossibleReturns)
				return nil, ("%q Unknown unit"):format(unit)
			end
			if isMod then
				local val = ModifierValues[mod]
				if val and ((val == "number" and lastPossibleReturns["string"]) or (val == "string" and lastPossibleReturns["number"])) then
					t[#t+1] = [[if type(value) == "]]
					t[#t+1] = val
					t[#t+1] = [[" ]]
					if unit and unit ~= "player" then
						t[#t+1] = "and UnitExists("
						t[#t+1] = ("%q"):format(unit)
						t[#t+1] = ") "
					end
					t[#t+1] = "then "
					if val == "number" then
						if lastPossibleReturns.string then
							finalPossibleReturns.string = true
						end
					else
						if lastPossibleReturns.number then
							finalPossibleReturns.number = true
						end
					end
				elseif lastPossibleReturns["nil"] then
					t[#t+1] = "if value "
					if unit and unit ~= "player" then
						t[#t+1] = "and UnitExists("
						t[#t+1] = ("%q"):format(unit)
						t[#t+1] = ") "
					end
					t[#t+1] = "then "
				elseif unit and unit ~= "player" then
					t[#t+1] = "if UnitExists("
					t[#t+1] = ("%q"):format(unit)
					t[#t+1] = ") "
					t[#t+1] = "then "
				else
					t[#t+1] = "do "
				end
			else
				t[#t+1] = "if not value "
				if unit and unit ~= "player" then
					t[#t+1] = "and UnitExists("
					t[#t+1] = ("%q"):format(unit)
					t[#t+1] = ") "
				end
				t[#t+1] = "then "
				ends = ends + 1
				
				if cachedTags[code[3]] then
					hasCache = cachedTags[code[3]]
					doneCachedTags[code[3]] = true
					if not firstCache then
						t[#t+1] = "if "
						t[#t+1] = hasCache
						t[#t+1] = " ~= NIL then value = "
						t[#t+1] = hasCache
						t[#t+1] = ";"
						t[#t+1] = " else "
					end
				elseif code[3]:find("^~") and cachedTags[code[3]:sub(2)] then
					hasCache = cachedTags[code[3]:sub(2)]
					if mod:find("^~") then
						mod = mod:sub(2)
						negate = not negate
						func = Tags[mod]
						assert(func)
					end
					doneCachedTags[code[3]:sub(2)] = true
					if not firstCache then
						t[#t+1] = "if "
						t[#t+1] = hasCache
						t[#t+1] = " ~= NIL then value = not "
						t[#t+1] = hasCache
						t[#t+1] = (" and %q;"):format(L["True"])
						t[#t+1] = " else "
					end
				end
			end
			if not negate or (isMod and ModifierReturns[mod]:find("nil")) or (not isMod and TagReturns[mod]:find("nil")) then
				if negate and isMod then
					t[#t+1] = "local old_value = value;"
				end
				local g
				if isMod then
					g = ModifierGlobals[mod]
				else
					g = TagGlobals[mod]
				end
				if g then
					g = newList((";"):split(g))
					for _,v in ipairs(g) do
						globals[v] = true
					end
					g = del(g)
				end
				if type(func) == "function" then
					local data = newList()
					data.isMod = isMod
					data.unit = unit
					data.unit_str = unit and ("(%q)"):format(unit) or "(unit)"
					data.arg = param
					data.arg_str = param and toliteral(param) or "(nil)"
					data.arg_string = param and ("(%q)"):format(param) or '("nil")'
					data.argtype = ("(%q)"):format(type(param))
					local hash = newDict('func', func(data))
					func = hash.func
					if hash.globals then
						g = newList((";"):split(hash.globals))
						for _,v in ipairs(g) do
							globals[v] = true
						end
						g = del(g)
					end
					hash = del(hash)
					data = del(data)
				end
				if param then
					if type(param) == "string" and param:find("%[") and param:find("%]") then
						local single = param:find("^%[[^%[%]]+%]$") and mod ~= "_"
						local possibleReturns
						if not single then
							if mod ~= "_" then
								t[#t+1] = "local v = value;"
							end

							t[#t+1] = "" -- placeholder

							local t_id = #t
							local madeT = false
							local t_numstart = math.random(0, 10000000)
							local t_num = t_numstart
							local t_literal = newList()
							local st = 1
							t[#t+1] = [[do ]]
							while st <= param:len() do
								local start, fin, left, literal = param:find("^(.-)(%b[])", st)
								if not start then
									local rest = param:sub(st)
									if rest:len() > 0 then
										if not madeT then
											t[#t+1] = ("value = %q"):format(rest)
											if possibleReturns then
												possibleReturns = del(possibleReturns)
											end
											possibleReturns = newSet("string")
										else
											t_literal[t_num+1] = rest
										end
									end
									break
								end
								if not madeT and (left:len() > 0 or start ~= 1 or fin ~= param:len()) then
									madeT = true
								end
								local inner = literal:sub(2, -2)
								st = fin+1
								if left:len() > 0 then
									t_literal[t_num] = left
								end

								local u = newList()
								if madeT then
									u[#u+1] = "do "
								end
								u[#u+1] = "value = nil;"

								local result, ret = compileTree(parse(splitCode(inner)), globals, cachedTags, doneCachedTags)
								u[#u+1] = result

								if madeT then
									t_num = t_num + 1
									u[#u+1] = ([[t_%d = value or '';]]):format(t_num)
									u[#u+1] = [[end;]]
								else
									possibleReturns = ret
								end
								local s = table_concat(u)
								u = del(u)
								t[#t+1] = s
							end
							if madeT then
								local u = newList()
								for i = t_numstart+1, t_num do
									u[i] = ("local t_%d = '';"):format(i)
								end
								t[t_id] = table_concat(u)
								u = del(u)
								t[#t+1] = [[value = ]]
								if t_literal[0] then
									t[#t+1] = ("%q"):format(t_literal[0])
									t[#t+1] = " .. "
								end
								for i = t_numstart+1, t_num+1 do
									if i <= t_num then
										t[#t+1] = "t_"
										t[#t+1] = i
										t[#t+1] = " .. "
									end
									if t_literal[i] then
										t[#t+1] = ("%q"):format(t_literal[i])
										t[#t+1] = " .. "
									end
								end
								t[#t] = [[;]]
								possibleReturns = newSet("string")
							end
							t_literal = del(t_literal)
							t[#t+1] = [[end;]]
						end
						if mod ~= "_" then
							if single then
								local result, ret = compileTree(parse(splitCode(param:sub(2, -2))), globals, cachedTags, doneCachedTags)
								result = result:gsub("%f[A-Za-z0-9_]value%f[^A-Za-z0-9_]", "arg")
								possibleReturns = ret
								t[#t+1] = "local arg;"
								t[#t+1] = result
							else
								t[#t+1] = "local arg = value;"
							end
							local set = newSet((";"):split((isMod and ModifierArguments[mod]) or (not isMod and TagArguments[mod]) or 'nil'))
							if set["nil"] then
								if set.number then
									if set.string then
										-- do nothing
									else
										if (not possibleReturns["nil"] and not possibleReturns.number) or possibleReturns.string then
											t[#t+1] = "arg = tonumber(arg);"
										end
									end
								else
									if set.string then
										if possibleReturns.number then
											if possibleReturns["nil"] then
												t[#t+1] = "if arg then arg = tostring(arg) end;"
											else
												t[#t+1] = "arg = tostring(arg);"
											end
										end
									else
										error(("Bad argument list. Can't just return nil. %q"):format(mod))
									end
								end
							else
								if set.number then
									if set.string then
										if possibleReturns["nil"] then
											t[#t+1] = "if not arg then arg = '' end;"
										end
									else
										if possibleReturns["nil"] or possibleReturns.string then
											t[#t+1] = "arg = tonumber(arg) or 0;"
										end
									end
								else
									if set.string then
										if not possibleReturns["nil"] then
											if possibleReturns.number then
												t[#t+1] = "arg = tostring(arg);"
											end
										else
											if possibleReturns.number then
												t[#t+1] = "if arg then arg = tostring(arg) else arg = '' end;"
											else
												t[#t+1] = "if not arg then arg = '' end;"
											end
										end
									else
										error(("Bad argument list. Can't just return nil. %q"):format(mod))
									end
								end
							end
							set = del(set)
							if not single then
								t[#t+1] = "value = v;"
							end
						end
						if possibleReturns then
							possibleReturns = del(possibleReturns)
						end
						func = func:gsub("%(${arg}%)", "(arg)")
						func = func:gsub("${arg}", "(arg)")
						local pos = func:find("${arg:string}")
						if pos and func:find("${arg:string}", pos+1) then
							t[#t+1] = "local arg_string = tostring(arg);"
							func = func:gsub("%(${arg:string}%)", "(arg_string)")
							func = func:gsub("${arg:string}", "(arg_string)")
						else
							func = func:gsub("%(${arg:string}%)", "(tostring(arg))")
							func = func:gsub("${arg:string}", "(tostring(arg))")
						end
						local pos = func:find("${argtype}")
						if pos and func:find("${argtype}", pos+1) then
							t[#t+1] = "local argtype = type(arg);"
							func = func:gsub("%(${argtype}%)", "(argtype)")
							func = func:gsub("${argtype}", "(argtype)")
						else
							func = func:gsub("%(${argtype}%)", "(type(arg))")
							func = func:gsub("${argtype}", "(type(arg))")
						end
					else
						func = func:gsub("%(${arg}%)", toliteral(param))
						func = func:gsub("${arg}", toliteral(param))
						func = func:gsub("%(${arg:string}%)", ("(%q)"):format(param))
						func = func:gsub("${arg:string}", ("(%q)"):format(param))
						func = func:gsub("%(${argtype}%)", ("(%q)"):format(type(param)))
						func = func:gsub("${argtype}", ("(%q)"):format(type(param)))
					end
				else
					func = func:gsub("%(${arg}%)", "(nil)")
					func = func:gsub("${arg}", "(nil)")
					func = func:gsub("%(${arg:string}%)", '("nil")')
					func = func:gsub("${arg:string}", '("nil")')
					func = func:gsub("%(${argtype}%)", '("nil")')
					func = func:gsub("${argtype}", '("nil")')
				end
				if unit then
					func = func:gsub("%(${unit}%)", ("(%q)"):format(unit))
					func = func:gsub("${unit}", ("(%q)"):format(unit))
				else
					func = func:gsub("%(${unit}%)", "(unit)")
					func = func:gsub("${unit}", "(unit)")
				end
				if mod ~= "_" then
					t[#t+1] = func
					if func ~= "" then
						t[#t+1] = ";"
					end
				end
				if hasCache then
					t[#t+1] = hasCache
					t[#t+1] = " = value;"
				end
				if hasCache and not firstCache then
					t[#t+1] = " end;"
				end
				if negate then
					t[#t+1] = "value = not value and "
					if isMod then
						t[#t+1] = "old_value"
						t[#t+1] = " end;"
					else
						t[#t+1] = ("%q;"):format(L["True"])
					end
					finalPossibleReturns["string"] = true
					finalPossibleReturns["nil"] = true
				elseif isMod then
					t[#t+1] = " end;"
				end
				if not negate then
					local rets = newSet((';'):split((isMod and ModifierReturns[mod]) or (not isMod and TagReturns[mod])))
					if rets.same then
						rets.same = nil
						if not lastPossibleReturns.number and not lastPossibleReturns.string then
							finalPossibleReturns.number = true
							finalPossibleReturns.string = true
						else
							finalPossibleReturns.number = lastPossibleReturns.number
							finalPossibleReturns.string = lastPossibleReturns.string
						end
					end
					for v in pairs(rets) do
						finalPossibleReturns[v] = true
					end
					if lastPossibleReturns["nil"] then
						finalPossibleReturns["nil"] = true
					end
					if lastPossibleReturns.number then
						finalPossibleReturns.number = true -- FIXME
					end
					if lastPossibleReturns.string then
						finalPossibleReturns.string = true -- FIXME
					end
					rets = del(rets)
				end
			else
				t[#t+1] = "value = nil;"
			end
		else
			t = del(t)
			code = del(code)
			lastPossibleReturns = del(lastPossibleReturns)
			finalPossibleReturns = del(finalPossibleReturns)
			return nil, ("%q Unknown modifier or tag"):format(mod)
		end
	else
		t = del(t)
		code = del(code)
		lastPossibleReturns = del(lastPossibleReturns)
		finalPossibleReturns = del(finalPossibleReturns)
		return nil, ("%q Unknown separator"):format(code[2])
	end
	code = del(code)
	lastPossibleReturns = del(lastPossibleReturns)
	local s = table_concat(t)
	t = del(t)
	return s, ends, finalPossibleReturns
end

function compileTree(code, globals, cachedTags, doneCachedTags)
	local madeDoneCachedTags = not doneCachedTags
	if madeDoneCachedTags then
		doneCachedTags = newList()
	end
	local q, e, possibleReturns = _compileTree(code, globals, cachedTags, doneCachedTags, newSet("nil"))
	if not q then
		return ("value = %q;"):format(e), newSet("string")
	end
	for i = 1, e do
		q = q .. "end;"
	end
	if madeDoneCachedTags then
		doneCachedTags = del(doneCachedTags)
	end
	return q, possibleReturns
end

local function select2(min, max, ...)
	if min <= max then
		return select(min, ...), select2(min+1, max, ...)
	end
end

function parse(...)
	local n = select('#', ...)
	if n == 1 then
		return newList((...))
	end
	local q_num = 0
	local q_pos
	local x_pos
	for i = 1, n do
		if select(i, ...) == "?" then
			q_num = q_num + 1
			if not q_pos then
				q_pos = i
			end
		elseif select(i, ...) == "!" then
			q_num = q_num - 1
			if q_num == 0 then
				x_pos = i
				break
			end
		end
	end
	if q_pos then
		if not x_pos then
			return newList(parse(select2(1, q_pos - 1, ...)), "?", parse(select(q_pos + 1, ...)))
		else
			return newList(parse(select2(1, q_pos - 1, ...)), "?", parse(select2(q_pos + 1, x_pos - 1, ...)), "!", parse(select(x_pos + 1, ...)))
		end
	end
	return newList(parse(select2(1, n - 2, ...)), ":", select(n, ...))
end

local function cleanText(text)
	while true do
		local lastText = text
		text = string_gsub(text, " +$", "")
		text = string_gsub(text, "^ +", "")
		text = string_gsub(text, "  +", " ")
		text = string_gsub(text, "|| ||", "||")
		text = string_gsub(text, "^||", "")
		text = string_gsub(text, "||$", "")
		if text == lastText then
			if text == "" then
				return " "
			end
			return text
		end
	end
end

local UnitToLocale = {player = L["Player"], target = L["Target"], pet = L["%s's pet"]:format(L["Player"]), focus = L["Focus-target"], mouseover = L["Mouse-over"]}
setmetatable(UnitToLocale, {__index=function(self, unit)
	if unit:find("pet$") then
		local nonPet = unit:sub(1, -4)
		self[unit] = L["%s's pet"]:format(self[nonPet])
		return self[unit]
	elseif not unit:find("target$") then
		if unit:find("^party%d$") then
			local num = unit:match("^party(%d)$")
			self[unit] = L["Party member #%d"]:format(num)
			return self[unit]
		elseif unit:find("^raid%d%d?$") then
			local num = unit:match("^raid(%d%d?)$")
			self[unit] = L["Raid member #%d"]:format(num)
			return self[unit]
		elseif unit:find("^partypet%d$") then
			local num = unit:match("^partypet(%d)$")
			self[unit] = UnitToLocale["party" .. num .. "pet"]
			return self[unit]
		elseif unit:find("^raidpet%d%d?$") then
			local num = unit:match("^raidpet(%d%d?)$")
			self[unit] = UnitToLocale["raid" .. num .. "pet"]
			return self[unit]
		end
		self[unit] = unit
		return unit
	end
	local nonTarget = unit:sub(1, -7)
	self[unit] = L["%s's target"]:format(self[nonTarget])
	return self[unit]
end})

local GetOfflineTime, GetAFKTime, GetDeadTime

local figureCachedTags
do
	local _figureCachedTags
	local function __figureCachedTags(code, data)
		for i = 1, #code, 2 do
			local v = code[i]
			if type(v) == "table" then
				__figureCachedTags(v, data)
			else
				local tag, param, negate, unit = splitParam(v)
				if param and param:find("%[") and param:find("%]") then
					_figureCachedTags(v, data)
				elseif Tags[tag] and tag ~= "Text" then
					data[v] = (data[v] or 0) + 1
				end
			end
		end
	end
	function _figureCachedTags(code, data)
		local st = 1
		while st <= code:len() do
			local start, fin, left, literal = code:find("^(.-)(%b[])", st)
			if not start then
				break
			end
			local inner = literal:sub(2, -2)
			st = fin+1
			local tree = parse(splitCode(inner))
			__figureCachedTags(tree, data)
		end
	end
	function figureCachedTags(code)
		local data = newList()
		_figureCachedTags(code, data)
		for k, v in pairs(data) do
			if k:find("^~") and data[k:sub(2)] then
				data[k:sub(2)] = data[k:sub(2)] + v
				data[k] = nil
			end
		end
		for k, v in pairs(data) do
			if v == 1 then
				data[k] = nil
			else
				data[k] = k:gsub("[^A-Za-z0-9_]", "_")
			end
		end
		return data
	end
end

local antiAlias
do
	local antiAliasWithUnit
	local handler__unit = nil
	local function handler(literal)
		local inner = literal:sub(2,-2)
		local code = newList(splitCode(inner))
		for i = 1, #code, 2 do
			local bit = code[i]
			if tonumber(bit) then
				code[i] = "Text(" .. bit .. ")"
			end
		end
		for _, opList in ipairs(ops) do
			for i = 2, #code-1, 2 do
				local op = code[i]
				if opList[op] then
					local nextOp_pos
					for j = i+2, #code-1, 2 do
						local nextOp = code[j]
						if opSet[op] then
							nextOp_pos = j
							break
						end
					end
					code[i] = ":" .. opList[op] .. "(["
					if nextOp_pos then
						table.insert(code, nextOp_pos, "])")
					else
						code[#code+1] = "])"
					end
					table.insert(code, 1, "[")
					code[#code+1] = "]"
					local text = table.concat(code)
					code = del(code)
					return handler(text)
				end
			end
		end
		local real = newList()
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			if not unit then
				unit = handler__unit
			end
			local func
			local isMod = false
			if code[i-1] ~= ":" then
				func = not Tags[bit] and TagAliases[bit]
			else
				func = not Modifiers[bit] and ModifierAliases[bit]
				if not func and not Modifiers[bit] then
					func = not Tags[bit] and TagAliases[bit]
				else
					isMod = true
				end
			end
			if i == 1 then
				real[1] = "["
			else
				real[#real+1] = code[i-1]
			end
			if negate then
				real[#real+1] = "~"
			end
			if type(param) == "string" and param:find("%[") and param:find("%]") then
				param = antiAliasWithUnit(param, unit)
			end
			if func then
				local group = not isMod and (func:find("[:!%?%+%-%*/^%%<>=]") or func:find("~="))
				if group then
					if negate then
						real[#real+1] = "Text(["
					else
						real[#real+1] = "_(["
					end
				end
				if param and (func:find("[%(:!%?%+%-%*/^%%<>=]") or func:find("~=")) then
					if func:find("%(arg%)") then
						real[#real+1] = func:gsub("%(arg%)", "(" .. param .. ")")
					else
						real[#real+1] = joinParam(bit, param, nil, unit)
					end
				elseif unit and func:find("#") then
					real[#real+1] = joinParam(bit, param, nil, unit)
				elseif not func:find("[:!%?%+%-%*/^%%<>=]") and not func:find("~=") then
					local fbit, fparam, fnegate, funit = splitParam(func)
					real[#real+1] = joinParam(fbit, antiAliasWithUnit(fparam, unit) or param, fnegate, funit or unit)
				elseif unit then
					if param then
						real[#real+1] = joinParam(bit, param, nil, unit)
					else
						local fcode = newList(splitCode(func))
						for j = 1, #fcode, 2 do
							if j > 1 then
								real[#real+1] = fcode[j-1]
							end
							local fbit, fparam, fnegate, funit = splitParam(fcode[j])
							real[#real+1] = joinParam(fbit, antiAliasWithUnit(fparam, unit), fnegate, unit)
						end
					end
				else
					real[#real+1] = func
					if param then
						real[#real+1] = "("
						real[#real+1] = param
						real[#real+1] = ")"
					end
				end
				if group then
					real[#real+1] = "])"
				end
			else
				real[#real+1] = joinParam(bit, param, nil, unit)
			end
		end
		code = del(code)

		real[#real+1] = ']'
		local str = table_concat(real):gsub("~~", "")
		real = del(real)
		return str
	end
	function antiAliasWithUnit(key, unit)
		if key == nil then
			return nil
		end
		if unit == nil then
			local value = rawget(antiAlias, key)
			if value then
				return value
			end
		end
		local value = key:gsub("~~", "")
		while true do
			local lastValue = value
			local last_handler__unit = handler__unit
			handler__unit = unit
			value = lastValue:gsub("(%b[])", handler)
			handler__unit = last_handler__unit
			if value == lastValue then
				break
			end
		end
		if unit == nil then
			antiAlias[key] = value
		end
		return value
	end
	antiAlias = setmetatable({}, {__index = function(self, key)
		return antiAliasWithUnit(key)
	end, __mode= 'k'})
end

function DogTag:UnaliasCode(code)
	return antiAlias[code]
end
precondition(DogTag, 'UnaliasCode', function(self, code)
	argCheck(code, 2, "string")
end)

local function enumLines(text)
	local lines = newList(('\n'):split(text))
	local t = newList()
	local indent = 0
	for i, v in ipairs(lines) do
		if v:match("end;$") or v:match("else$") or v:match("^ *elseif") then
			indent = indent - 1
		end
		for j = 1, indent do
			t[#t+1] = "    "
		end
		t[#t+1] = v
		t[#t+1] = " -- "
		t[#t+1] = i
		t[#t+1] = "\n"
		if v:match("then$") or v:match("do$") or v:match("else$") then
			indent = indent + 1
		end
	end
	lines = del(lines)
	local s = table_concat(t)
	t = del(t)
	return s
end

function DogTag:CreateFunctionFromCode(code, notDebug)
	if type(code) ~= "string" then
		error(("Bad argument #2 to `CreateFunctionFromCode'. Expected %q, got %q."):format("string", type(code)), 2)
	end
	code = antiAlias[code]
	if cleanText(code) == "" then
		return ""
	end
	local cachedTags = figureCachedTags(code)
	local t = newList()
	t[#t+1] = ([[local DogTag = LibStub(%q);]]):format(MAJOR_VERSION)
	t[#t+1] = [[local colors = DogTag.__colors;]]
	t[#t+1] = "" -- placeholder
	t[#t+1] = [[local NIL = {};]]
	t[#t+1] = [[local cleanText = DogTag.__cleanText;]]
	t[#t+1] = [[return function(unit) ]]
	local u = newList()
	for k, v in pairs(cachedTags) do
		u[#u+1] = "local "
		u[#u+1] = v
		u[#u+1] = " = NIL;"
	end
	t[#t+1] = table_concat(u)
	t[#t+1] = "local opacity;"
	u = del(u)
	t[#t+1] = [[if not UnitExists(unit) then return ]]
	local globals = newList()
	globals['table.concat'] = true
	globals['string.gsub'] = true
	globals['type'] = true
	globals['tonumber'] = true
	globals['UnitExists'] = true
	if code:find("[%[:!%?%+%-%*/^%%<>=~][Nn][Aa][Mm][Ee][%]:!%?%+%-%*/^%%<>=~]") then
		t[#t+1] = [=[DogTag___UnitToLocale[unit]]=]
		globals['DogTag.__UnitToLocale'] = true
	else
		t[#t+1] = [[""]]
	end
	t[#t+1] = [[; end;]]
	t[#t+1] = "" -- placeholder
	local tmp_id = #t
	local madeTmp = false
	local tmp_num = 0
	local tmp_literal = newList()
	local st = 1
	while st <= code:len() do
		local start, fin, left, literal = code:find("^(.-)(%b[])", st)
		if not start then
			local rest = code:sub(st)
			if rest:len() > 0 then
				if not madeTmp then
					t[#t+1] = ("return %q"):format(rest)
				else
					tmp_literal[tmp_num+1] = rest
				end
			end
			break
		end
		if not madeTmp and (left:len() > 0 or start ~= 1 or fin ~= code:len()) then
			madeTmp = true
		end
		local inner = literal:sub(2, -2)
		st = fin+1
		if left:len() > 0 then
			tmp_literal[tmp_num] = left
		end
		if st < code:len() then
			madeTmp = true
		end

		local u = newList()
		u[#u+1] = "do local value;"

		local result, ret = compileTree(parse(splitCode(inner)), globals, cachedTags)
		if ret then
			ret = del(ret)
		end
		u[#u+1] = result
		if not madeTmp then
			u[#u+1] = [[return cleanText(value or ''), opacity;]]
		else
			tmp_num = tmp_num + 1
			u[#u+1] = ([[tmp_%d = value or '';]]):format(tmp_num)
		end
		u[#u+1] = [[end;]]
		local s = table_concat(u)
		u = del(u)
		t[#t+1] = s
	end
	local u = newList()
	for i = 1, tmp_num do
		u[i] = ("local tmp_%d = '';"):format(i)
	end
	t[tmp_id] = table_concat(u)
	u = del(u)
	local u = newList()
	for k in pairs(globals) do
		if k:find("^Lib[A-Z][A-Za-z%-]%-%d+%.%d+$") then
			u[#u+1] = "local "
			u[#u+1] = k:gsub("%-.*", "")
			u[#u+1] = "Lib = Rock('"
			u[#u+1] = k
			u[#u+1] = "', false, true);"
		elseif k:find("[A-Za-z%-]*%-%d%.%d") then
			if Rock then
				Rock(k, false, true) -- try to load
			end
			if AceLibrary then
				AceLibrary:HasInstance(k) -- try to load
			end
			if LibStub(k, true) then
				-- catches Rock and AceLibrary
				u[#u+1] = "local "
				u[#u+1] = k:gsub("%-.*", "")
				u[#u+1] = "Lib = LibStub('"
				u[#u+1] = k
				u[#u+1] = "');"
			elseif AceLibrary then
				-- non-LibStub AceLibrary libs.
				u[#u+1] = "local "
				u[#u+1] = k:gsub("%-.*", "")
				u[#u+1] = "Lib = AceLibrary and AceLibrary:HasInstance('"
				u[#u+1] = k
				u[#u+1] = "') and AceLibrary('"
				u[#u+1] = k
				u[#u+1] = "');"
			end
		else
			u[#u+1] = "local "
			u[#u+1] = k:gsub("%.", "_")
			u[#u+1] = " = "
			u[#u+1] = k
			u[#u+1] = ";"
		end
	end
	globals = del(globals)
	t[3] = table_concat(u)
	u = del(u)
	if tmp_num > 0 then
		t[#t+1] = [[return cleanText(]]
		if tmp_literal[0] then
			t[#t+1] = ("%q"):format(tmp_literal[0])
			t[#t+1] = " .. "
		end
		for i = 1, tmp_num+1 do
			if i <= tmp_num then
				t[#t+1] = [[tmp_]]
				t[#t+1] = i
				t[#t+1] = " .. "
			end
			if tmp_literal[i] then
				t[#t+1] = ("%q"):format(tmp_literal[i])
				t[#t+1] = " .. "
			end
		end
		t[#t] = [[), opacity;]]
	end
	tmp_literal = del(tmp_literal)
	t[#t+1] = [[end;]]
	cachedTags = del(cachedTags)
	local s = table_concat(t)
	t = del(t)
	if not notDebug then
		s = enumLines(s:gsub(";", ";\n"):gsub("%f[A-Za-z_]do%f[^A-Za-z_]", "do\n"):gsub("%f[A-Za-z_]then%f[^A-Za-z_]", "then\n"):gsub("%f[A-Za-z_]else%f[^A-Za-z_]", "else\n"):gsub("\n *", "\n")) -- avoid interning the new string if not debugging
	end
	return s
end
precondition(DogTag, 'CreateFunctionFromCode', function(self, code, notDebug)
	argCheck(code, 2, "string")
	argCheck(notDebug, 3, "boolean", "nil")
end)

local FakeGlobals = {}

local codeToFunction; codeToFunction = setmetatable({}, {__index=function(self, key)
	local code = antiAlias[key]
	if code ~= key then
		self[key] = self[code]
		return self[key]
	end
	local s = DogTag:CreateFunctionFromCode(code, true)
	DogTag.__colors = colors
	DogTag.__cleanText = cleanText
	DogTag.__UnitToLocale = UnitToLocale
	DogTag.__currentAuras = currentAuras
	DogTag.__currentDebuffTypes = currentDebuffTypes
	DogTag.__currentNumDebuffs = currentNumDebuffs
	DogTag.__guildNotes = guildNotes
	DogTag.__officerNotes = officerNotes
	DogTag.__GetOfflineTime = GetOfflineTime
	DogTag.__GetAFKTime = GetAFKTime
	DogTag.__GetDeadTime = GetDeadTime
	DogTag.__FigureNPCGuild = FigureNPCGuild
	DogTag.__FigureZone = FigureZone
	DogTag.__FigureFaction = FigureFaction
	DogTag.__talentSpecs = talentSpecs
	DogTag.__talentSpecNames = talentSpecNames
	DogTag.__addonVersions = addonVersions
	DogTag.__better_UnitLevel = better_UnitLevel
	DogTag.__castData = castData
	for k, v in pairs(FakeGlobals) do
		DogTag[k] = v
	end
	local func, err = loadstring(s)
	local val
	if not func then
		geterrorhandler()(("DogTag-2.0: Error (%s) loading code %q. Please inform ckknight."):format(err, code))
		val = codeToFunction[""]
	else
		local status, result = pcall(func)
		if not status then
			geterrorhandler()(("DogTag-2.0: Error (%s) running code %q. Please inform ckknight."):format(result, code))
			val = codeToFunction[""]
		else
			val = result
		end
	end
	DogTag.__colors = nil
	DogTag.__cleanText = nil
	DogTag.__UnitToLocale = nil
	DogTag.__currentAuras = nil
	DogTag.__currentDebuffTypes = nil
	DogTag.__currentNumDebuffs = nil
	DogTag.__guildNotes = nil
	DogTag.__officerNotes = nil
	DogTag.__GetOfflineTime = nil
	DogTag.__GetAFKTime = nil
	DogTag.__GetDeadTime = nil
	DogTag.__FigureNPCGuild = nil
	DogTag.__FigureZone = nil
	DogTag.__FigureFaction = nil
	DogTag.__talentSpecs = nil
	DogTag.__talentSpecNames = nil
	DogTag.__addonVersions = nil
	DogTag.__better_UnitLevel = nil
	DogTag.__castData = nil
	for k in pairs(FakeGlobals) do
		DogTag[k] = nil
	end
	self[code] = val
	return val
end, __mode='k'})

local NormalUnits = {player = true, pet = true, target = true, mouseover = true, focus = true}
for i = 1, 4 do
	NormalUnits['party' .. i] = true
	NormalUnits['partypet' .. i] = true
end
for i = 1, 40 do
	NormalUnits['raid' .. i] = true
end

local codeToEventList; codeToEventList = setmetatable({}, {__index = function(self, realkey)
	local key = antiAlias[realkey]
	if key ~= realkey then
		self[realkey] = self[key]
		return self[realkey]
	end
	local t = newList()
	local len = key:len()
	local st = 1
	while st <= len do
		local start, fin, left, literal = key:find("^(.-)(%b[])", st)
		if not start then
			break
		end
		local inner = literal:sub(2, -2)
		st = fin+1
		local code = newList(splitCode(inner))
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			local func
			if code[i-1] ~= ":" then
				func = Tags[bit]
			else
				func = Modifiers[bit] or Tags[bit]
			end
			local isMod = func == Modifiers[bit]
			local eventList = (isMod and ModifierEvents[bit]) or (not isMod and TagEvents[bit])
			if type(func) == "function" then
				local data = newList()
				data.isMod = isMod
				data.unit = unit
				data.unit_str = unit and ("(%q)"):format(unit) or "(unit)"
				data.arg = param
				data.arg_str = param and toliteral(param) or "(nil)"
				data.arg_string = param and ("(%q)"):format(param) or '("nil")'
				data.argtype = ("(%q)"):format(type(param))
				local hash = newDict('func', func(data))
				data = del(data)
				if hash.events then
					if eventList then
						eventList = eventList .. ";" .. hash.events
					else
						eventList = hash.events
					end
				end
				hash = del(hash)
			end
			if unit == true then
				unit = nil
			end
			if unit and not NormalUnits[unit] then
				if eventList then
					eventList = eventList .. ";" .. "Update"
				else
					eventList = "Update"
				end
			end
			if eventList then
				local events = newList((";"):split(eventList))
				for _,event in ipairs(events) do
					if event == "Special_IsUnit" then
						if param == "player" or param == "pet" or param == "target" or param == "focus" or param == "mouseover" or param:find("^party%d$") or param:find("^partypet%d$") or param:find("^raid%d+$") or param:find("^raidpet%d+$") then
							t[event .. "=" .. param] = true
						else
							t["Update=" .. param] = true
						end
						if unit then
							t[event .. "=" .. unit] = true
						else
							t[event] = true
						end
					else
						local e, u = event:match("([A-Z_]+)%(([a-z]+)%)")
						if not e then
							e, u = event, unit
						end
						if u then
							t[e .. "=" .. u] = true
						else
							t[e] = true
						end
					end
				end
				events = del(events)
			end
			if type(param) == "string" and param:find("%[") and param:find("%]") then
				for k, v in pairs(codeToEventList[param]) do
					t[k] = v
				end
			end
		end
		code = del(code)
	end
	self[key] = t
	return t
end, __mode = 'k'})

local eventData = setmetatable({}, {__index = function(self, event)
	local t = newList()
	self[event] = t
	if event:find("^[A-Z_]+$") then
		DogTag.frame:RegisterEvent(event)
	end
	return t
end})
local buckets = setmetatable({}, {__index = function(self, event)
	local t = newList()
	self[event] = t
	return t
end})

local registry
local codeHash

local frameToText = {}
local textToFrame = {}

local toUpdate = {}

local function clamp(value, min, max)
	if value < min then
		return min
	elseif value > max then
		return max
	else
		return value
	end
end

local function updateText(text, code, unit)
	local success, ret, alpha = pcall(codeToFunction[code], unit)
	if success then
		toUpdate[text] = nil
		text:SetText(ret)
		if alpha then
			alpha = tonumber(alpha) or 1
			text:SetAlpha(clamp(alpha, 0, 1))
		end
	else
		geterrorhandler()(("%s.%d: Error with code %q on %q. %s"):format(MAJOR_VERSION, MINOR_VERSION, code, unit, ret))
	end
end

local runUpdate

local hasReputationEvent = false
local hasThreatEvent = false
local hasAuraEvent = false
local hasMouseoverEvent = false
local hasAltKeyEvent = false
local hasCtrlKeyEvent = false
local hasShiftKeyEvent = false
local hasFastUpdateEvent = false
local hasUpdateEvent = false
local hasSlowUpdateEvent = false
local hasTalentEvent = false
local hasAddonVersionEvent = false
local hasAFKTimeEvent = false
local hasDeadTimeEvent = false
local hasOfflineTimeEvent = false
local hasCastEvent = false
local shouldRecheckEventData = false
local function RecheckEventData()
	shouldRecheckEventData = false
	hasReputationEvent = not not rawget(eventData, "Reputation")
	hasThreatEvent = not not rawget(eventData, "Threat")
	hasMouseoverEvent = not not rawget(eventData, "Mouseover")
	hasFastUpdateEvent = not not rawget(eventData, "FastUpdate")
	hasUpdateEvent = not not rawget(eventData, "Update")
	hasSlowUpdateEvent = not not rawget(eventData, "SlowUpdate")
	hasAltKeyEvent = not not rawget(eventData, "AltKey")
	hasCtrlKeyEvent = not not rawget(eventData, "CtrlKey")
	hasShiftKeyEvent = not not rawget(eventData, "ShiftKey")
	hasAFKTimeEvent = not not rawget(eventData, "AFKTime")
	hasDeadTimeEvent = not not rawget(eventData, "DeadTime")
	hasOfflineTimeEvent = not not rawget(eventData, "OfflineTime")
	hasTalentEvent = not not rawget(eventData, "Talent")
	hasAddonVersionEvent = not not rawget(eventData, "AddonVersion")
	hasCastEvent = not not rawget(eventData, "Cast")
	for k in pairs(eventData) do
		if not k:find("^[A-Z_]+$") then
			if k ~= "Reputation" and k ~= "Mouseover" and k ~= "Threat" and k ~= "FastUpdate" and k ~= "Update" and k ~= "SlowUpdate" and k ~= "AltKey" and k ~= "ShiftKey" and k ~= "CtrlKey" and k ~= "AFKTime" and k ~= "DeadTime" and k ~= "OfflineTime" and k ~= "Talent" and k ~= "AddonVersion" and k ~= "Cast" then
				hasAuraEvent = true
				break
			end
		end
	end
end

function DogTag:AddFontString(text, frame, unit, code)
	if not IsLegitimateUnit[unit] then
		code = ""
	end
	local outline = ""
	if code:find("^%[[Oo][Uu][Tt][Ll][Ii][Nn][Ee]%]") then
		code = code:sub(10)
		outline = "OUTLINE"
	elseif code:find("^%[[Tt][Hh][Ii][Cc][Kk][Oo][Uu][Tt][Ll][Ii][Nn][Ee]%]") then
		code = code:sub(15)
		outline = "OUTLINE, THICKOUTLINE"
	end
	if codeHash[text] then
		if codeHash[text] == code and registry[text] == unit and textToFrame[text] == frame and select(3, text:GetFont()) == outline then
			toUpdate[text] = true
			return
		end
		self:RemoveFontString(text)
	end
	codeHash[text] = code
	registry[text] = unit
	if NormalUnits[unit] then
		local events = codeToEventList[code]

		for event in pairs(events) do
			local event, u = ("="):split(event, 2)
			if not u then
				u = unit
			end
			if eventData[event][text] and eventData[event][text] ~= u then
				if type(eventData[event][text]) == "string" then
					local old = eventData[event][text]
					eventData[event][text] = newList()
					eventData[event][text][old] = true
				end
				eventData[event][text][u] = true
			else
				eventData[event][text] = u
			end
		end
	end
	registry[text] = unit

	shouldRecheckEventData = true

	if not frameToText[frame] then
		frameToText[frame] = newList()
	end
	frameToText[frame][text] = unit
	textToFrame[text] = frame

	local font, fontsize = text:GetFont()
	text:SetFont(font, fontsize, outline)
	DogTag.__isMouseOver = textToFrame[text] == lastMouseFocus
	updateText(text, codeHash[text], unit)
end
precondition(DogTag, 'AddFontString', function(self, text, frame, unit, code)
	argCheck(text, 2, "table")
	argCheck(frame, 3, "table")
	argCheck(unit, 4, "string")
	argCheck(code, 5, "string")
end)

local alwaysEvents = {
	PARTY_MEMBERS_CHANGED = true,
	PLAYER_ENTERING_WORLD = true,
	PLAYER_TARGET_CHANGED = true,
	PLAYER_FOCUS_CHANGED = true,
	PLAYER_PET_CHANGED = true,
	UPDATE_MOUSEOVER_UNIT = true,
	UNIT_PET = true,
	ADDON_LOADED = true,
	PLAYER_LOGIN = true,
	UPDATE_FACTION = true,
	UNIT_AURA = true,
	MODIFIER_STATE_CHANGED = true,
	PLAYER_FLAGS_CHANGED = true,
	GUILD_ROSTER_UPDATE = true,
	PLAYER_GUILD_UPDATE = true,
	INSPECT_TALENT_READY = true,
	UNIT_SPELLCAST_SENT = true,
	UNIT_SPELLCAST_START = true,
	UNIT_SPELLCAST_CHANNEL_START = true,
	UNIT_SPELLCAST_STOP = true,
	UNIT_SPELLCAST_FAILED = true,
	UNIT_SPELLCAST_DELAYED = true,
	UNIT_SPELLCAST_CHANNEL_UPDATE = true,
	UNIT_SPELLCAST_CHANNEL_STOP = true,
	UNIT_SPELLCAST_INTERRUPTED = true,
	UNIT_SPELLCAST_CHANNEL_INTERRUPTED = true,
}

function DogTag:RemoveFontString(text)
	local code = codeHash[text]
	if not code then
		return
	end

	local unit = registry[text]
	registry[text] = nil
	codeHash[text] = nil
	if NormalUnits[unit] then
		local events = codeToEventList[code]

		for event in pairs(events) do
			local t = eventData[event]
			if type(t[text]) == "table" then
				del(t[text])
			end
			t[text] = nil
			if not next(t) then
				eventData[event] = del(t)
				if event:find("^[A-Z_]+$") then
					if not alwaysEvents[event] then
						self.frame:UnregisterEvent(event)
					end
					buckets[event] = del(buckets[event])
				end
			end
		end

		shouldRecheckEventData = true
	end

	local frame = textToFrame[text]
	textToFrame[text] = nil
	frameToText[frame][text] = nil
	if not next(frameToText[frame]) then
		frameToText[frame] = del(frameToText[frame])
	end
end
precondition(DogTag, 'RemoveFontString', function(self, text)
	argCheck(text, 2, "table")
end)

function DogTag:UpdateFontString(text)
	local unit = registry[text]

	if unit then
		DogTag.__isMouseOver = textToFrame[text] == lastMouseFocus
		updateText(text, codeHash[text], unit)
	end
end
precondition(DogTag, 'UpdateFontString', function(self, text)
	argCheck(text, 2, "table")
end)

local auraQueue = {}

function DogTag:UpdateAllForUnit(unit)
	if not IsLegitimateUnit[unit] then
		return
	end

	if rawget(currentAuras, unit) then
		currentAuras[unit] = del(currentAuras[unit])
		currentDebuffTypes[unit] = del(currentDebuffTypes[unit])
		currentNumDebuffs[unit] = nil
		auraQueue[unit] = true
	end
	
	if rawget(castData, unit) then
		castData[unit] = del(castData[unit])
	end

	for text, u in pairs(registry) do
		if u == unit then
			toUpdate[text] = true
		end
	end
	for event, data in pairs(eventData) do
		for text, u in pairs(data) do
			if u == unit or data[u] then
				toUpdate[text] = true
			end
		end
	end

	runUpdate()
end
precondition(DogTag, 'UpdateAllForUnit', function(self, unit)
	argCheck(unit, 2, "string")
end)

function DogTag:UpdateAllForFrame(frame)
	if not frame:IsShown() then
		return
	end

	local list = frameToText[frame]
	if not list then
		return
	end

	for text, unit in pairs(list) do
		toUpdate[text] = true
	end

	runUpdate()
end
precondition(DogTag, 'UpdateAllForFrame', function(self, frame)
	argCheck(frame, 2, "table")
end)

function DogTag:Evaluate(unit, code)
	if not IsLegitimateUnit[unit] then
		return ""
	end
	DogTag.__isMouseOver = false

	local success, ret = pcall(codeToFunction[code], unit)
	if success then
		return ret
	else
		geterrorhandler()(("%s.%d: Error with code %q on %q. %s"):format(MAJOR_VERSION, MINOR_VERSION, code, unit, ret))
	end
end
precondition(DogTag, 'Evaluate', function(self, unit, code)
	argCheck(unit, 2, "string")
	argCheck(code, 3, "string")
end)

do
	local FixCasing__external = false
	local function handler(literal)
		local inner = literal:sub(2, -2)
		local code = newList(splitCode(inner))
		if FixCasing__external then
			for i = 1, #code, 2 do
				local op = code[i-1]
				local bit, param, negate, unit = splitParam(code[i])
				local negated_bit = negate and "~" .. bit or bit
				if op == ":" then
					for _,opList in ipairs(ops) do
						for op_, mod in pairs(opList) do
							if mod == negated_bit then
								local t = newList()
								t[#t+1] = "["
								for j = 1, i-3 do
									t[#t+1] = code[j]
								end
								t[#t+1] = "["
								t[#t+1] = code[i-2]
								t[#t+1] = op_
								t[#t+1] = joinParam("Text", param, negate, unit)
								t[#t+1] = "]"
								for j = i+1, #code do
									t[#t+1] = code[j]
								end
								t[#t+1] = "]"
								local result = table.concat(t)
								code = del(code)
								t = del(t)
								return handler(result)
							end
						end
					end
				elseif negated_bit == "Text" then
					if tonumber(param) then
						code[i] = param
					else
						if param:match("^(%b[])$") then
							code[i] = joinParam("_", param, negate, unit)
						end
					end
				end
			end
		end
		local t = newList()
		for i = 1, #code, 2 do
			local bit, param, negate, unit = splitParam(code[i])
			
			if i == 1 then
				t[1] = "["
			else
				local isMod = code[i-1] == ':'
				if not isMod then
					t[#t+1] = ' '
				end
				t[#t+1] = code[i-1]
				if not isMod then
					t[#t+1] = ' '
				end
			end
			
			t[#t+1] = joinParam(bit, param, negate, unit)
		end
		code = del(code)
		t[#t+1] = ']'
		local str = table_concat(t):gsub("~~", "")
		t = del(t)
		return str
	end
	
	function FixCasing(code, external)
		code = code:gsub("|c%x%x%x%x%x%x%x%x", ""):gsub("|r", "")
		local outline = ""
		if code:find("^%[[Oo][Uu][Tt][Ll][Ii][Nn][Ee]%]") then
			code = code:sub(10)
			outline = "[Outline]"
		elseif code:find("^%[[Tt][Hh][Ii][Cc][Kk][Oo][Uu][Tt][Ll][Ii][Nn][Ee]%]") then
			code = code:sub(15)
			outline = "[ThickOutline]"
		end
		local prev = FixCasing__external
		FixCasing__external = external or prev
		local result = outline .. (code:gsub("(%b[])", handler))
		FixCasing__external = prev
		return result
	end
	
	function DogTag:FixCasing(code)
		return FixCasing(code, true)
	end
	precondition(DogTag, 'FixCasing', function(self, code)
		argCheck(code, 2, "string")
	end)
end

do
	local colors = {
		tag = "00ffff", -- cyan
		number = "ff00ff", -- fushcia
		modifier = "00ff00", -- green
		literal = "ff7f7f", -- pink
		operator = "7f7fff", -- light blue
		unit = "ff0000", -- red
		result = "ffffff", -- white
	}
	
	local function handler(literal)
		local inner = literal:sub(2, -2)
		local t = newList()
		local code = newList(splitCode(inner))
		for i = 1, #code, 2 do
			t[#t+1] = "|cff"
			t[#t+1] = colors.operator
			if i == 1 then
				t[#t+1] = "["
			else
				local isMod = code[i-1] == ":"
				if not isMod then
					t[#t+1] = " "
				end
				t[#t+1] = code[i-1]
				if not isMod then
					t[#t+1] = " "
				end
			end	
			t[#t+1] = "|r"
			local bit, param, negate, unit = splitParam(code[i])
			if tonumber(bit) then
				-- Number
				t[#t+1] = "|cff"
				t[#t+1] = colors.number
			elseif i == 1 or code[i-1] ~= ":" then
				-- Tag
				t[#t+1] = "|cff"
				t[#t+1] = colors.tag
			else
				-- Modifier?
				local proper_bit = correctCasing[bit]
				if (Tags[proper_bit] or TagAliases[proper_bit]) and not Modifiers[proper_bit] and not ModifierAliases[proper_bit] then
					t[#t+1] = "|cff"
					t[#t+1] = colors.tag
				else
					t[#t+1] = "|cff"
					t[#t+1] = colors.modifier
				end
			end
			if bit ~= "_" or not param then
				if negate then
					t[#t+1] = "~"
				end
				t[#t+1] = bit
				t[#t+1] = "|r"
				if unit then
					t[#t+1] = "|cff"
					t[#t+1] = colors.operator
					t[#t+1] = "#"
					t[#t+1] = "|r"
					t[#t+1] = "|cff"
					t[#t+1] = colors.unit
					t[#t+1] = unit
					t[#t+1] = "|r"
				end
				if param then
					t[#t+1] = "|cff"
					t[#t+1] = colors.operator
					t[#t+1] = "("
					t[#t+1] = "|r"
					if tonumber(param) then
						t[#t+1] = "|cff"
						t[#t+1] = colors.number
						t[#t+1] = param
						t[#t+1] = "|r"
					else
						t[#t+1] = DogTag:ColorizeTag(param)
					end
					t[#t+1] = "|cff"
					t[#t+1] = colors.operator
					t[#t+1] = ")"
					t[#t+1] = "|r"
				end
			else	
				if tonumber(param) then
					t[#t+1] = "|cff"
					t[#t+1] = colors.number
					t[#t+1] = param
					t[#t+1] = "|r"
				else
					t[#t+1] = DogTag:ColorizeTag(param)
				end
			end
		end
		t[#t+1] = "|cff"
		t[#t+1] = colors.operator
		t[#t+1] = "]"
		t[#t+1] = "|r"
		local result = table.concat(t)
		t = del(t)
		return result
	end
	function DogTag:ColorizeTag(code)
		if not code then
			return nil
		end
		code = code:gsub("|c%x%x%x%x%x%x%x%x", ""):gsub("|r", "")
		local t = newList()
		
		local last = 1
		while true do
			local alpha, bravo, literal = code:find("(%b[])", last)
			if not alpha then
				break
			end
			if alpha ~= last then
				t[#t+1] = "|cff"
				t[#t+1] = colors.literal
				t[#t+1] = code:sub(last, alpha-1)
				t[#t+1] = "|r"
			end
			t[#t+1] = handler(literal)
			last = bravo + 1
		end
		
		if last <= code:len() then
			t[#t+1] = "|cff"
			t[#t+1] = colors.literal
			t[#t+1] = code:sub(last)
			t[#t+1] = "|r"
		end
		
		local result = table.concat(t)
		t = del(t)
		return result
	end
	precondition(DogTag, 'ColorizeTag', function(self, code)
		argCheck(code, 2, "string", "nil")
	end)
end

function DogTag:UNIT_AURA(unit)
	auraQueue[unit] = true
end

function DogTag:UNIT_SPELLCAST_SENT(unit, spell, rank, target)
	if not hasCastEvent then
		return
	end
	if target and target ~= "" then
		castData[unit].nextSpell = spell
		castData[unit].nextRank = rank and tonumber(rank:match("%d+"))
		castData[unit].nextTarget = target
	end
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end

function DogTag:UNIT_SPELLCAST_START(unit)
	if not hasCastEvent then
		return
	end
	if rawget(castData, unit) then
		local nextSpell = castData[unit].nextSpell
		local nextRank = castData[unit].nextRank
		local nextTarget = castData[unit].nextTarget
		castData[unit] = del(castData[unit])
		if castData[unit].spell == nextSpell and castData[unit].rank == nextRank then
			castData[unit].target = nextTarget
		end
	end
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end
DogTag.UNIT_SPELLCAST_CHANNEL_START = DogTag.UNIT_SPELLCAST_START

function DogTag:UNIT_SPELLCAST_STOP(unit)
	if not hasCastEvent or not rawget(castData, unit) or not castData[unit].casting or castData[unit].fading then
		return
	end
	castData[unit].stopTime = GetTime()
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end

function DogTag:UNIT_SPELLCAST_FAILED(unit)
	if not hasCastEvent or not rawget(castData, unit) or castData[unit].fading then
		return
	end
	castData[unit].stopTime = GetTime()
	castData[unit].stopMessage = _G.FAILED
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end

function DogTag:UNIT_SPELLCAST_INTERRUPTED(unit)
	if not hasCastEvent or not rawget(castData, unit) then
		return
	end
	castData[unit].stopTime = GetTime()
	castData[unit].stopMessage = _G.INTERRUPTED
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end
DogTag.UNIT_SPELLCAST_CHANNEL_INTERRUPTED = DogTag.UNIT_SPELLCAST_INTERRUPTED

function DogTag:UNIT_SPELLCAST_DELAYED(unit)
	if not hasCastEvent or not rawget(castData, unit) or not castData[unit].casting or castData[unit].stopTime then
		return
	end
	
	local spell, rank, displayName, icon, startTime, endTime = UnitCastingInfo(unit)

	if not spell or not startTime or not endTime then
		return
	end
	
	local oldStart = castData[unit].startTime
	
	startTime = startTime / 1000
	castData[unit].startTime = startTime
	castData[unit].endTime = endTime / 1000

	castData[unit].delay = castData[unit].delay + (startTime - oldStart)
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end

function DogTag:UNIT_SPELLCAST_CHANNEL_UPDATE(unit)
	if not hasCastEvent or not rawget(castData, unit) or castData[unit].casting or castData[unit].stopTime then
		return
	end
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
	
	local spell, rank, displayName, icon, startTime, endTime = UnitChannelInfo(unit)
	if not spell then
		for k,v in pairs(castData[unit]) do
			castData[unit][k] = nil
		end
		return
	end
	
	local oldStart = castData[unit].startTime
	startTime = startTime / 1000
	castData[unit].startTime = startTime
	castData[unit].endTime = endTime / 1000
	castData[unit].delay = castData[unit].delay + (oldStart - startTime)
end

function DogTag:UNIT_SPELLCAST_CHANNEL_STOP(unit)
	if not hasCastEvent or not rawget(castData, unit) or castData[unit].casting or castData[unit].stopTime then
		return
	end
	
	castData[unit].stopTime = GetTime()
	
	for text, u in pairs(eventData.Cast) do
		if unit == u then
			toUpdate[text] = true
		end
	end
end

local offlineTimes = {}
local afkTimes = {}
local deadTimes = {}

function DogTag:PARTY_MEMBERS_CHANGED()
	for k, v in pairs(currentAuras) do
		currentAuras[k] = del(v)
		currentDebuffTypes[k] = del(currentDebuffTypes[k])
		currentNumDebuffs[k] = nil
	end
	for text, unit in pairs(registry) do
		auraQueue[unit] = true
		toUpdate[text] = true
	end

	local unitType, max = "raid", GetNumRaidMembers()
	if max == 0 then
		unitType, max = "party", GetNumPartyMembers()
	end

	for i = 0, max do
		local unit = unitType .. i
		if i == 0 then
			unit = "player"
		end
		if not UnitExists(unit) then
			break
		end
		local name = GetNameServer(UnitName(unit))
		if not UnitIsConnected(unit) then
			if not offlineTimes[name] then
				offlineTimes[name] = time()
			end
			afkTimes[name] = nil
		else
			offlineTimes[name] = nil
			if UnitIsAFK(unit) then
				if not afkTimes[name] then
					afkTimes[name] = time()
				end
			else
				afkTimes[name] = nil
			end
		end
		if UnitIsDeadOrGhost(unit) then
			if not deadTimes[name] then
				deadTimes[name] = time()
			end
		else
			deadTimes[name] = nil
		end
	end

	runUpdate()
end
DogTag.PLAYER_ENTERING_WORLD = DogTag.PARTY_MEMBERS_CHANGED
DogTag.UNIT_PET = DogTag.PARTY_MEMBERS_CHANGED

function DogTag:UPDATE_FACTION()
	for i = 1, GetNumFactions() do
		local name = GetFactionInfo(i)
		factionList[name] = true
	end
end

function DogTag:PLAYER_TARGET_CHANGED()
	self:UpdateAllForUnit('target')
end

function DogTag:PLAYER_FOCUS_CHANGED()
	self:UpdateAllForUnit('focus')
end

function DogTag:PLAYER_PET_CHANGED()
	self:UpdateAllForUnit('pet')
end

local inMouseover = false
function DogTag:UPDATE_MOUSEOVER_UNIT()
	inMouseover = true
	self:UpdateAllForUnit('mouseover')
end

local function clearCodes()
	for k in pairs(codeToFunction) do
		codeToFunction[k] = nil
	end
	for k, v in pairs(codeToEventList) do
		codeToEventList[k] = del(v)
	end
	for k in pairs(antiAlias) do
		antiAlias[k] = nil
	end
	for k, v in pairs(eventData) do
		eventData[k] = del(v)
	end
	for k, v in pairs(buckets) do
		buckets[k] = del(v)
	end
	for k, v in pairs(TagAliasFunctions) do
		TagAliases[k] = v()
	end
	for k, v in pairs(ModifierAliasFunctions) do
		ModifierAliases[k] = v()
	end

	local self = DogTag
	self.frame:UnregisterAllEvents()
	for k in pairs(alwaysEvents) do
		self.frame:RegisterEvent(k)
	end

	DogTag:PARTY_MEMBERS_CHANGED()
end

local AddonFinders = {}
function DogTag:ADDON_LOADED()
	-- for extra addons' help, such as MobHealth3 or MobInfo2
	AceLibrary = _G.AceLibrary
	local good = false
	for v in pairs(AddonFinders) do
		local kind, name, func = v[1], v[2], v[3]
		if kind == "_G" then
			if _G[name] then
				func(_G[name])
				AddonFinders[v] = nil
				good = true
			end
		elseif kind == "AceLibrary" then
			if AceLibrary and AceLibrary:HasInstance(name) then
				func(AceLibrary(name))
				AddonFinders[v] = nil
				good = true
			end
		elseif kind == "LibStub" then
			if LibStub:GetLibrary(name, true) then
				func(LibStub:GetLibrary(name))
				AddonFinders[v] = nil
				good = true
			end
		elseif kind == "Rock" then
			if Rock:HasLibrary(name) then
				func(Rock:GetLibrary(name))
				AddonFinders[v] = nil
				good = true
			end
		end
	end
	if not good then
		return
	end
	clearCodes()
	return true
end

function DogTag:PLAYER_LOGIN()
	if not self:ADDON_LOADED() then
		self:PARTY_MEMBERS_CHANGED()
	end
	self:UPDATE_FACTION()

	self:PLAYER_FLAGS_CHANGED("player")
end

function DogTag:SetColorConstantTable(t)
	colors = t
	-- need to update the colors reference in the functions
	for k in pairs(codeToFunction) do
		codeToFunction[k] = nil
	end
	clearCodes()
end
precondition(DogTag, 'SetColorConstantTable', function(self, t)
	argCheck(t, 2, "table")
end)

function DogTag:MODIFIER_STATE_CHANGED(key)
	local eventName
	if key == "ALT" then
		eventName = "AltKey"
		if not hasAltKeyEvent then
			return
		end
	elseif key == "CTRL" then
		eventName = "CtrlKey"
		if not hasCtrlKeyEvent then
			return
		end
	elseif key == "SHIFT" then
		eventName = "ShiftKey"
		if not hasShiftKeyEvent then
			return
		end
	else
		return
	end

	for text in pairs(eventData[eventName]) do
		if text:IsVisible() and text:GetNumPoints() > 0 then
			DogTag.__isMouseOver = textToFrame[text] == lastMouseFocus
			updateText(text, codeHash[text], registry[text])
		end
	end
end

function DogTag:OnMouseoverUpdate()
	local toUpdate = newList()
	local mouseFocus = GetMouseFocus()
	local list, list2 = frameToText[lastMouseFocus], frameToText[mouseFocus]
	lastMouseFocus = mouseFocus
	if hasMouseoverEvent then
		if list then
			if list2 then
				for text, unit in pairs(eventData.Mouseover) do
					if list[text] or list2[text] then
						toUpdate[text] = true
					end
				end
			else
				for text, unit in pairs(eventData.Mouseover) do
					if list[text] then
						toUpdate[text] = true
					end
				end
			end
		elseif list2 then
			for text, unit in pairs(eventData.Mouseover) do
				if list2[text] then
					toUpdate[text] = true
				end
			end
		end
	end

	for text, u in pairs(registry) do
		if u == "mouseover" or u.mouseover then
			toUpdate[text] = true
		end
	end
	for event, data in pairs(eventData) do
		for text, u in pairs(data) do
			if u == "mouseover" or u.mouseover then
				toUpdate[text] = true
			end
		end
	end

	for text in pairs(toUpdate) do
		if text:IsVisible() and text:GetNumPoints() > 0 then
			DogTag.__isMouseOver = textToFrame[text] == lastMouseFocus
			updateText(text, codeHash[text], registry[text])
		end
	end
	toUpdate = del(toUpdate)
end

function DogTag:OnReputationUpdate()
	if not hasReputationEvent then
		return
	end
	for text, unit in pairs(eventData.Reputation) do
		toUpdate[text] = true
	end
end

function DogTag:Threat_ThreatUpdated(unitName, unitID, targetHash, threatValue)
	if not Threat or not hasThreatEvent then
		return
	end
	for text, unit in pairs(eventData.Threat) do
		if not toUpdate[text] then
			local name = UnitName(unit)
			if (unit == unitID or unit[unitID] or (name and Threat:UnitIsUnit(name, targetHash))) then
				toUpdate[text] = true
			end
		end
	end
end

function GetAFKTime(unit)
	local name = GetNameServer(UnitName(unit))
	local t = afkTimes[name]
	if not t then
		return nil
	end
	local time_diff = time() - t
	if time_diff < 3600 then
		return ("%d:%02d"):format(time_diff/60, time_diff%60)
	else
		return ("%d:%02d:%02d"):format(time_diff/3600, time_diff%3600/60, time_diff%60)
	end
end

function GetDeadTime(unit)
	local name = GetNameServer(UnitName(unit))
	local t = deadTimes[name]
	if not t then
		return nil
	end
	local time_diff = time() - t
	if time_diff < 3600 then
		return ("%d:%02d"):format(time_diff/60, time_diff%60)
	else
		return ("%d:%02d:%02d"):format(time_diff/3600, time_diff%3600/60, time_diff%60)
	end
end

function GetOfflineTime(unit)
	local name = GetNameServer(UnitName(unit))
	local t = offlineTimes[name]
	if not t then
		return nil
	end
	local time_diff = time() - t
	if time_diff < 3600 then
		return ("%d:%02d"):format(time_diff/60, time_diff%60)
	else
		return ("%d:%02d:%02d"):format(time_diff/3600, time_diff%3600/60, time_diff%60)
	end
end

function DogTag:PLAYER_FLAGS_CHANGED(unit)
	buckets.PLAYER_FLAGS_CHANGED[unit or false] = true

	local name, server = UnitName(unit)
	if not name then
		return
	end
	if server and server ~= "" then
		name = name .. '-' .. server
	end
	local isAFK = UnitIsAFK(unit)

	if isAFK then
		if afkTimes[name] then
			return
		end
		afkTimes[name] = time()
	else
		if not afkTimes[name] then
			return
		end
		afkTimes[name] = nil
	end

	for text, u in pairs(eventData.AFKTime) do
		if not toUpdate[text] then
			local n, server = UnitName(unit)
			if n then
				if server and server ~= "" then
					n = n .. '-' .. server
				end
				if n == name then
					toUpdate[text] = true
				end
			end
		end
	end
end

local nextGuildRosterUpdate = 0
function DogTag:GUILD_ROSTER_UPDATE()
	buckets.GUILD_ROSTER_UPDATE[false] = true

	refreshGuildNotes(true)

	nextGuildRosterUpdate = GetTime() + 20
end

function DogTag:PLAYER_GUILD_UPDATE()
	refreshGuildNotes(true)
end

local function getTalentSpecName(class, nums)
	class = class:sub(1, 1) .. class:sub(2):lower()
	if nums[1] == 0 and nums[2] == 0 and nums[3] == 0 then
		return _G.NONE or "None"
	else
		local first, second, third
		if nums[1] >= nums[2] then
			if nums[1] >= nums[3] then
				first = 1
				if nums[2] >= nums[3] then
					second = 2
					third = 3
				else
					second = 3
					third = 2
				end
			else
				first = 3
				second = 1
				third = 2
			end
		else
			if nums[2] >= nums[3] then
				first = 2
				if nums[1] >= nums[3] then
					second = 1
					third = 3
				else
					second = 3
					third = 1
				end
			else
				first = 3
				second = 2
				third = 1
			end
		end
		local first_num = nums[first]
		local second_num = nums[second]
		if first_num*3/4 <= second_num then
			if first_num*3/4 <= nums[third] then
				-- all three balanced out
				return L["Hybrid"]
			else
				return L[class .. "_Tree_" .. first] .. "/" .. L[class .. "_Tree_" .. second]
			end
		else
			return L[class .. "_Tree_" .. first]
		end
	end
end

local nextInspectionCheck = 0
local function checkInspectionQueue()
	if not next(inspectionQueue) then
		return
	end
	if InspectFrame and InspectFrame.unit then
		return
	end
	local name = GetNameServer(UnitName("player"))
	if inspectionQueue[name] then
		NotifyInspect("player")
		return
	end
	
	local found
	local name = UnitExists("mouseover") and UnitIsPlayer("mouseover") and GetNameServer(UnitName("mouseover"))
	if inspectionQueue[name] then
		found = true
	 	if UnitIsVisible("mouseover") and UnitIsConnected("mouseover") then
			NotifyInspect("mouseover")
			return
		end
	end

	local name = UnitExists("target") and UnitIsPlayer("target") and GetNameServer(UnitName("target"))
	if inspectionQueue[name] then
		found = true
	 	if UnitIsVisible("target") and UnitIsConnected("target") then
			NotifyInspect("target")
			return
		end
	end

	local name = UnitExists("focus") and UnitIsPlayer("focus") and GetNameServer(UnitName("focus"))
	if inspectionQueue[name] then
		found = true
	 	if UnitIsVisible("focus") and UnitIsConnected("focus") then
			NotifyInspect("focus")
			return
		end
	end

	local numRaid = GetNumRaidMembers()
	if numRaid > 0 then
		for i = 1, numRaid do
			local unit = "raid" .. i
			local name = UnitExists(unit) and GetNameServer(UnitName(unit))
			if inspectionQueue[name] then
				found = true
				if UnitIsVisible(unit) and UnitIsConnected(unit) then
					NotifyInspect(unit)
					return
				end
			end
		end
		for i = 1, numRaid do
			local unit = "raid" .. i .. "target"
			local name = UnitExists(unit) and UnitIsPlayer(unit) and GetNameServer(UnitName(unit))
			if inspectionQueue[name] then
				found = true
				if UnitIsVisible(unit) and UnitIsConnected(unit) then
					NotifyInspect(unit)
					return
				end
			end
		end
	else
		local numParty = GetNumPartyMembers()
		for i = 1, numParty do
			local unit = "party" .. i
			local name = UnitExists(unit) and GetNameServer(UnitName(unit))
			if inspectionQueue[name] then
				found = true
			 	if UnitIsVisible(unit) and UnitIsConnected(unit) then
					NotifyInspect(unit)
					return
				end
			end
		end
		for i = 1, numParty do
			local unit = "party" .. i .. "target"
			local name = UnitExists(unit) and UnitIsPlayer(unit) and GetNameServer(UnitName(unit))
			if inspectionQueue[name] then
				found = true
			 	if UnitIsVisible(unit) and UnitIsConnected(unit) then
					NotifyInspect(unit)
					return
				end
			end
		end
	end
	
	if not found then
		for k in pairs(inspectionQueue) do
			inspectionQueue[k] = nil
		end
	end
end

function DogTag:INSPECT_TALENT_READY()
	local unit = InspectFrame and InspectFrame.unit
	local name, class
	if not unit then
		if not lastInspectName or GetTime() - lastInspectTime > 15 then
			return
		end
		name = lastInspectName
		class = lastInspectClass
	else
		if not UnitExists(unit) or not UnitIsPlayer(unit) or not UnitIsVisible(unit) or not UnitIsConnected(unit) then
			return
		end
		name = GetNameServer(UnitName(unit))
		local _
		_, class = UnitClass(unit)
	end
	lastInspectName = nil
	inspectionQueue[name] = nil
	
	local nums = newList()
	for i = 1, GetNumTalentTabs(true) do
		local total = 0
		for j = 1, GetNumTalents(i, true) do
			total = total + select(5, GetTalentInfo(i, j, true))
		end
		nums[i] = total
	end
	
	talentSpecs[name] = ("%d/%d/%d"):format(nums[1], nums[2], nums[3])
	talentSpecNames[name] = getTalentSpecName(class, nums)
	nums = del(nums)
	
	-- trigger "Talent" event
	if hasTalentEvent then
		for text, unit in pairs(eventData.Talent) do
			toUpdate[text] = true
		end
	end
	
	checkInspectionQueue()
end

function talentReceptor(sender, data)
	local nums = newList()
	for i,v in ipairs(data) do
		local num = 0
		for j,u in ipairs(v) do
			num = num + u
		end
		nums[i] = num
	end
	inspectionQueue[sender] = nil
	talentSpecs[sender] = ("%d/%d/%d"):format(nums[1], nums[2], nums[3])
	talentSpecNames[sender] = getTalentSpecName(data.class, nums)
	nums = del(nums)
	-- trigger "Talent" event
	if not hasTalentEvent then
		return
	end
	for text, unit in pairs(eventData.Talent) do
		toUpdate[text] = true
	end
end

function addonVersionReceptor(sender, addon, version)
	addonVersions[addon][sender] = version or _G.NONE or "None"
	-- trigger "Talent" event
	if not hasAddonVersionEvent then
		return
	end
	for text, unit in pairs(eventData.AddonVersion) do
		toUpdate[text] = true
	end
end

local namesChecked = {}
local num = 0
local nextTime = GetTime()
function runUpdate()
	num = num + 1
	local currentTime = GetTime()
	if currentTime > nextGuildRosterUpdate then
		if IsInGuild() then
			GuildRoster()
		end
		nextGuildRosterUpdate = currentTime + 20
	end
	if currentTime > nextInspectionCheck then
		nextInspectionCheck = currentTime + 30
		checkInspectionQueue()
	end
	if currentTime > nextTime then
		nextTime = nextTime + 1
		if hasAFKTimeEvent then
			for text in pairs(eventData.AFKTime) do
				local unit = registry[text]
				local name = GetNameServer(UnitName(unit))
				if name then
					if UnitIsVisible(unit) then
						if afkTimes[name] then
							toUpdate[text] = true
						end
					else
						if UnitIsAFK(unit) then
							if not afkTimes[name] then
								afkTimes[name] = time()
							end
							toUpdate[text] = true
						else
							if afkTimes[name] then
								afkTimes[name] = nil
								toUpdate[text] = true
							end
						end
					end
				end
			end
		end
		if hasOfflineTimeEvent then
			for text in pairs(eventData.OfflineTime) do
				if not toUpdate[text] then
					local unit = registry[text]
					local name, server = GetNameServer(UnitName(unit))
					if name then
						if offlineTimes[name] then
							toUpdate[text] = true
						end
					end
				end
			end
		end
		if hasDeadTimeEvent then
			for text in pairs(eventData.DeadTime) do
				local unit = registry[text]
				if UnitInRaid(unit) or UnitInParty(unit) then
					local name = GetNameServer(UnitName(unit))
					if name then
						if not namesChecked[name] then
							if UnitIsDeadOrGhost(unit) then
								if not deadTimes[name] then
									deadTimes[name] = time()
								end
								namesChecked[name] = true
								toUpdate[text] = true
							else
								if deadTimes[name] then
									deadTimes[name] = nil
									namesChecked[name] = true
									toUpdate[text] = true
								end
							end
						else
							toUpdate[text] = true
						end
					end
				elseif UnitIsDeadOrGhost(unit) then
					toUpdate[text] = true
				end
			end
			for k in pairs(namesChecked) do
				namesChecked[k] = nil
			end
		end
	end
	if num%5 == 0 and hasAuraEvent then
		if num%20 == 0 then
			for unit, v in pairs(currentAuras) do
				if not NormalUnits[unit] then
					currentAuras[unit] = del(v)
					currentDebuffTypes[unit] = del(currentDebuffTypes[unit])
					currentNumDebuffs[unit] = nil
				end
			end
		end
		for unit in pairs(auraQueue) do
			auraQueue[unit] = nil
			local t = newList()
			local u = newList()
			for i = 1, 40 do
				local name = UnitBuff(unit, i)
				if not name then
					break
				end
				t[name] = (t[name] or 0) + 1
			end
			local numDebuffs = 0
			local isFriend = UnitIsFriend("player", unit)
			for i = 1, 40 do
				local name,_,_,_,dispelType = UnitDebuff(unit, i)
				if not name then
					break
				end
				numDebuffs = numDebuffs + 1
				t[name] = (t[name] or 0) + 1
				if isFriend and dispelType then
					u[dispelType] = true
				end
			end
			local old = rawget(currentAuras, unit) or newList()
			local oldType = rawget(currentDebuffTypes, unit) or newList()
			local changed = newList()
			local changedDebuffTypes = newList()
			for k in pairs(t) do
				if not old[k] then
					changed[k] = true
				else
					old[k] = nil
				end
			end
			for k in pairs(old) do
				if not t[k] then
					changed[k] = true
				end
			end
			for k in pairs(u) do
				if not oldType[k] then
					changedDebuffTypes[k] = true
				else
					oldType[k] = nil
				end
			end
			for k in pairs(oldType) do
				if not u[k] then
					changedDebuffTypes[k] = true
				end
			end
			currentAuras[unit] = t
			currentDebuffTypes[unit] = u
			local oldNumDebuffs = currentNumDebuffs[unit]
			currentNumDebuffs[unit] = numDebuffs
			if oldNumDebuffs ~= numDebuffs then
				if rawget(eventData, 'NumDebuffs') then
					for text in pairs(eventData.NumDebuffs) do
						toUpdate[text] = true
					end
				end
			end
			old = del(old)
			oldType = del(oldType)
			for name in pairs(changed) do
				if rawget(eventData, name) then
					for text in pairs(eventData[name]) do
						toUpdate[text] = true
					end
				end
			end
			for dispelType in pairs(changedDebuffTypes) do
				if rawget(eventData, dispelType) then
					for text in pairs(eventData[dispelType]) do
						toUpdate[text] = true
					end
				end
			end
			changed = del(changed)
			changedDebuffTypes = del(changedDebuffTypes)
		end
	end
	for unit, data in pairs(castData) do
		if not IsNormalUnit[unit] then
			castData[unit] = del(data)
		elseif not data.stopTime and data.endTime and currentTime > data.endTime then
			if data.casting then
			 	if not UnitIsUnit("player", unit) then
					data.stopTime = currentTime
				end
			else
				data.stopTime = currentTime
			end
		end
	end
	if hasFastUpdateEvent then
		for text in pairs(eventData.FastUpdate) do
			toUpdate[text] = true
		end
	end
	if num%3 == 0 and hasUpdateEvent then
		for text in pairs(eventData.Update) do
			toUpdate[text] = true
		end
	end
	if num%200 == 0 and hasSlowUpdateEvent then
		for text in pairs(eventData.SlowUpdate) do
			toUpdate[text] = true
		end
	end
	local mouseoverIsTarget = UnitIsUnit("target", "mouseover")
	for event, units in pairs(buckets) do
		if units[false] then
			for text in pairs(eventData[event]) do
				toUpdate[text] = true
			end
		else
			if mouseoverIsTarget and not units.mouseover then
				units.mouseover = units.target
			end
			for text, unit in pairs(eventData[event]) do
				if not toUpdate[text] then
					if units[unit] then
						toUpdate[text] = true
					elseif type(unit) == "table" then
						for u in pairs(unit) do
							if units[u] then
								toUpdate[text] = true
							end
						end
					end
				end
			end
		end
		buckets[event] = del(units)
	end
	local noMouseover = not UnitExists("mouseover")
	if inMouseover and noMouseover then
		inMouseover = false
		for event, data in pairs(eventData) do
			for text, u in pairs(data) do
				if u == "mouseover" or u.mouseover then
					toUpdate[text] = true
				end
			end
		end
	end
	if noMouseover then
		for text in pairs(toUpdate) do
			if registry[text] == "mouseover" then
				toUpdate[text] = nil
			end
		end
	end
	for text in pairs(toUpdate) do
		toUpdate[text] = nil
		local unit = registry[text]
		if unit and text:IsVisible() and text:GetNumPoints() > 0 then
			DogTag.__isMouseOver = textToFrame[text] == lastMouseFocus
			updateText(text, codeHash[text], unit)
		end
	end
end

DogTag.frame = oldLib and oldLib.frame or CreateFrame("Frame", MAJOR_VERSION .. "_Frame")
local frame = DogTag.frame
frame:UnregisterAllEvents()
for k in pairs(alwaysEvents) do
	frame:RegisterEvent(k)
end
frame:SetScript("OnEvent", function(this, event, arg1, ...)
	local func = DogTag[event]
	if func then
		func(DogTag, arg1, ...)
		return
	end
	buckets[event][arg1 or false] = true
end)

local timePassed = 0
local firstTime = true
frame:SetScript("OnUpdate", function(this, elapsed)
	timePassed = timePassed + elapsed
	if timePassed > 0.05 then
		if firstTime then
			firstTime = false
			DogTag:PARTY_MEMBERS_CHANGED()
		end
		if shouldRecheckEventData then
			RecheckEventData()
		end
		timePassed = 0
		runUpdate()
	end
end)

if not oldLib or not oldLib.hookedRep then
	hooksecurefunc("ReputationWatchBar_Update", function()
		DogTag:OnReputationUpdate()
	end)
end
DogTag.hookedRep = true

registry = {}
DogTag.registry = registry
codeHash = {}
DogTag.codeHash = codeHash

if oldLib then
	for text, unit in pairs(oldLib.registry) do
		DogTag:AddFontString(text, unit, oldLib.codeHash[text])
	end
end

local helpFrame
function DogTag:OpenHelp()
	if not helpFrame then
		helpFrame = CreateFrame("Frame", MAJOR_VERSION .. "_HelpFrame", UIParent)
		helpFrame:SetWidth(600)
		helpFrame:SetHeight(300)
		helpFrame:SetPoint("CENTER", UIParent, "CENTER")
		helpFrame:EnableMouse(true)
		helpFrame:SetMovable(true)
		helpFrame:SetResizable(true)
		helpFrame:SetMinResize(600, 300)
		helpFrame:SetFrameLevel(50)
		helpFrame:SetFrameStrata("FULLSCREEN_DIALOG")

		local bg = newDict(
			'bgFile', [[Interface\DialogFrame\UI-DialogBox-Background]],
			'edgeFile', [[Interface\DialogFrame\UI-DialogBox-Border]],
			'tile', true,
			'tileSize', 32,
			'edgeSize', 32,
			'insets', newDict(
				'left', 5,
				'right', 6,
				'top', 5,
				'bottom', 6
			)
		)
		helpFrame:SetBackdrop(bg)
		bg.insets = del(bg.insets)
		bg = del(bg)
		helpFrame:SetBackdropColor(0, 0, 0)
		helpFrame:SetClampedToScreen(true)
		
		local header = CreateFrame("Frame", helpFrame:GetName() .. "_Header", helpFrame)
		helpFrame.header = header
		header:SetHeight(34.56)
		header:SetClampedToScreen(true)
		local left = header:CreateTexture(header:GetName() .. "_TextureLeft", "ARTWORK")
		header.left = left
		left:SetPoint("TOPLEFT")
		left:SetPoint("BOTTOMLEFT")
		left:SetWidth(11.54)
		left:SetTexture([[Interface\DialogFrame\UI-DialogBox-Header]])
		left:SetTexCoord(0.235, 0.28, 0.04, 0.58)
		local right = header:CreateTexture(header:GetName() .. "_TextureRight", "ARTWORK")
		header.right = right
		right:SetPoint("TOPRIGHT")
		right:SetPoint("BOTTOMRIGHT")
		right:SetWidth(11.54)
		right:SetTexture([[Interface\DialogFrame\UI-DialogBox-Header]])
		right:SetTexCoord(0.715, 0.76, 0.04, 0.58)
		local center = header:CreateTexture(header:GetName() .. "_TextureCenter", "ARTWORK")
		header.center = center
		center:SetPoint("TOPLEFT", left, "TOPRIGHT")
		center:SetPoint("BOTTOMRIGHT", right, "BOTTOMLEFT")
		center:SetTexture([[Interface\DialogFrame\UI-DialogBox-Header]])
		center:SetTexCoord(0.28, 0.715, 0.04, 0.58)
		
		local closeButton = CreateFrame("Button", helpFrame:GetName() .. "_CloseButton", helpFrame, "UIPanelCloseButton")
		helpFrame.closeButton = closeButton
		closeButton:SetFrameLevel(helpFrame:GetFrameLevel()+5)
		closeButton:SetScript("OnClick", function(this)
			this:GetParent():Hide()
		end)
		closeButton:SetPoint("TOPRIGHT", helpFrame, "TOPRIGHT", -5, -5)
		
		header:EnableMouse(true)
		header:RegisterForDrag("LeftButton")
		header:SetScript("OnDragStart", function(this)
			isDragging = true
			this:GetParent():StartMoving()
		end)
		header:SetScript("OnDragStop", function(this)
			isDragging = false
			this:GetParent():StopMovingOrSizing()
		end)

		local titleText = header:CreateFontString(header:GetName() .. "_FontString", "OVERLAY", "GameFontNormal")
		helpFrame.titleText = titleText
		titleText:SetText(L["DogTag Help"])
		titleText:SetPoint("CENTER", helpFrame, "TOP", 0, -8)
		titleText:SetHeight(26)
		titleText:SetShadowColor(0, 0, 0)
		titleText:SetShadowOffset(1, -1)

		header:SetPoint("LEFT", titleText, "LEFT", -32, 0)
		header:SetPoint("RIGHT", titleText, "RIGHT", 32, 0)
		
		local sizer_se = CreateFrame("Frame", helpFrame:GetName() .. "_SizerSoutheast", helpFrame)
		helpFrame.sizer_se = sizer_se
		sizer_se:SetPoint("BOTTOMRIGHT", helpFrame, "BOTTOMRIGHT", 0, 0)
		sizer_se:SetWidth(25)
		sizer_se:SetHeight(25)
		sizer_se:EnableMouse(true)
		sizer_se:RegisterForDrag("LeftButton")
		sizer_se:SetScript("OnDragStart", function(this)
			isDragging = true
			this:GetParent():StartSizing("BOTTOMRIGHT")
		end)
		sizer_se:SetScript("OnDragStop", function(this)
			isDragging = false
			this:GetParent():StopMovingOrSizing()
		end)
		local line1 = sizer_se:CreateTexture(sizer_se:GetName() .. "_Line1", "BACKGROUND")
		line1:SetWidth(14)
		line1:SetHeight(14)
		line1:SetPoint("BOTTOMRIGHT", -10, 10)
		line1:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
		local x = 0.1 * 14/17
		line1:SetTexCoord(1/32 - x, 0.5, 1/32, 0.5 + x, 1/32, 0.5 - x, 1/32 + x, 0.5)

		local line2 = sizer_se:CreateTexture(sizer_se:GetName() .. "_Line2", "BACKGROUND")
		line2:SetWidth(11)
		line2:SetHeight(11)
		line2:SetPoint("BOTTOMRIGHT", -10, 10)
		line2:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
		local x = 0.1 * 11/17
		line2:SetTexCoord(1/32 - x, 0.5, 1/32, 0.5 + x, 1/32, 0.5 - x, 1/32 + x, 0.5)

		local line3 = sizer_se:CreateTexture(sizer_se:GetName() .. "_Line3", "BACKGROUND")
		line3:SetWidth(8)
		line3:SetHeight(8)
		line3:SetPoint("BOTTOMRIGHT", -10, 10)
		line3:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
		local x = 0.1 * 8/17
		line3:SetTexCoord(1/32 - x, 0.5, 1/32, 0.5 + x, 1/32, 0.5 - x, 1/32 + x, 0.5)
		
		local mainPane = CreateFrame("Frame", helpFrame:GetName() .. "_MainPane", helpFrame)
		helpFrame.mainPane = mainPane
		local bg = newDict(
			'bgFile', [[Interface\Buttons\WHITE8X8]],
			'edgeFile', [[Interface\Tooltips\UI-Tooltip-Border]],
			'tile', true,
			'tileSize', 16,
			'edgeSize', 16,
			'insets', newDict(
				'left', 3,
				'right', 3,
				'top', 3,
				'bottom', 3
			)
		)
		mainPane:SetBackdrop(bg)
		bg.insets = del(bg.insets)
		bg = del(bg)
		mainPane:SetBackdropBorderColor(0.6, 0.6, 0.6)
		mainPane:SetBackdropColor(0, 0, 0)
		mainPane:SetPoint("TOPLEFT", helpFrame, "TOPLEFT", 12, -35)
		mainPane:SetPoint("BOTTOMRIGHT", helpFrame, "BOTTOMRIGHT", -12, 30)
		
		local scrollFrame = CreateFrame("ScrollFrame", mainPane:GetName() .. "_ScrollFrame", mainPane)
		local scrollChild = CreateFrame("Frame", mainPane:GetName() .. "_ScrollChild", scrollFrame)
		local scrollBar = CreateFrame("Slider", mainPane:GetName() .. "_ScrollBar", scrollFrame, "UIPanelScrollBarTemplate")
		mainPane.scrollFrame = scrollFrame
		mainPane.scrollChild = scrollChild
		mainPane.scrollBar = scrollBar

		scrollFrame:SetScrollChild(scrollChild)
		scrollFrame:SetPoint("TOPLEFT", mainPane, "TOPLEFT", 9, -9)
		scrollFrame:SetPoint("BOTTOMRIGHT", mainPane, "BOTTOMRIGHT", -28, 12)
		scrollFrame:EnableMouseWheel(true)
		scrollFrame:SetScript("OnMouseWheel", function(this, change)
			local childHeight = scrollChild:CalculateHeight()
			local frameHeight = scrollFrame:GetHeight()
			if childHeight <= frameHeight then
				return
			end

			nextFreeScroll = GetTime() + 1

			local diff = childHeight - frameHeight

			local delta = 1
			if change > 0 then
				delta = -1
			end

			local value = scrollBar:GetValue() + delta*24/diff
			if value < 0 then
				value = 0
			elseif value > 1 then
				value = 1
			end
			scrollBar:SetValue(value) -- will trigger OnValueChanged
		end)

		scrollChild:SetHeight(10)
		scrollChild:SetWidth(10)

		local first = true
		function scrollChild:CalculateHeight()
			local html = self.html
			local t = newList(html:GetRegions())
			local top, bottom
			for i,v in ipairs(t) do
				if v:GetTop() and (not top or top < v:GetTop()) then
					top = v:GetTop()
				end
				if v:GetBottom() and (not bottom or bottom > v:GetBottom()) then
					bottom = v:GetBottom()
				end
			end
			t = del(t)
			return top and (top - bottom) or 10
		end
		_G.scrollChild = scrollChild

		scrollBar:SetPoint("TOPLEFT", scrollFrame, "TOPRIGHT", 0, -16)
		scrollBar:SetPoint("BOTTOMLEFT", scrollFrame, "BOTTOMRIGHT", 0, 16)
		scrollBar:SetMinMaxValues(0, 1)
		scrollBar:SetValueStep(1e-5)
		scrollBar:SetValue(0)
		scrollBar:SetWidth(16)
		scrollBar:SetScript("OnValueChanged", function(this)
			local max = scrollChild:CalculateHeight() - scrollFrame:GetHeight()

			local val = scrollBar:GetValue() * max
			
			if math.abs(scrollFrame:GetVerticalScroll() - val) < 1 then
				return
			end

			scrollFrame:SetVerticalScroll(val)

			scrollFrame:UpdateScrollChildRect()
		end)
		scrollBar:EnableMouseWheel(true)
		scrollBar:SetScript("OnMouseWheel", function(this, ...)
			scrollFrame:GetScript("OnMouseWheel")(scrollFrame, ...)
		end)
		
		local html = CreateFrame("SimpleHTML", scrollChild:GetName() .. "_HTML", scrollChild)
		scrollChild.html = html
		html:SetFontObject('p', GameFontNormal)
		html:SetSpacing('p', 3)
		
		html:SetFontObject('h1', GameFontHighlightLarge)
		local font, height, flags = GameFontHighlightLarge:GetFont()
		height = height * 1.25
		html:SetFont('h1', font, height, flags)
		html:SetSpacing('h1', height/2)
		
		html:SetFontObject('h2', GameFontHighlightLarge)
		local _, height = GameFontHighlightLarge:GetFont()
		html:SetSpacing('h2', height/2)
		
		html:SetFontObject('h3', GameFontHighlightNormal)
		local _, height = GameFontHighlightLarge:GetFont()
		html:SetSpacing('h3', height/2)
		
		html:SetHeight(1)
		html:SetWidth(400)
		html:SetPoint("TOPLEFT", 0, 0)
		html:SetJustifyH("LEFT")
		html:SetJustifyV("TOP")
		
		scrollFrame:SetScript("OnSizeChanged", function(this)
			html:SetWidth(this:GetWidth())
			html:SetText(html.text)
		end)
		
		local function _fix__handler(text)
			if text:sub(1, 2) == "{{" and text:sub(-2) == "}}" then
				if text:len() == 5 and not text:sub(3, 3):find("[A-Za-z0-9_]") then
					if text:sub(3, 3) == "~" then
						local x = "[~One]"
						x = DogTag:ColorizeTag(x)
						x = x:match("^|cff%x%x%x%x%x%x%[|r(|cff%x%x%x%x%x%x~One|r)|cff%x%x%x%x%x%x%]|r$")
						return x:gsub("One", "")
					end
					local x = "[One" .. text:sub(3, 3) .. "Two]"
					x = DogTag:ColorizeTag(x)
					return x:match("^|cff%x%x%x%x%x%x%[|r|cff%x%x%x%x%x%xOne|r(.*)|cff%x%x%x%x%x%xTwo|r|cff%x%x%x%x%x%x%]|r$")
				elseif text:sub(3, 3) == ":" or text:sub(3, 3) == "#" then
					local x = "[___" .. text:sub(3, -3) .. "]"
					x = DogTag:ColorizeTag(x)
					return x:match("^|cff%x%x%x%x%x%x%[|r|cff%x%x%x%x%x%x___|r(.*)|cff%x%x%x%x%x%x%]|r$")
				else
					local x = "[" .. text:sub(3, -3) .. "]"
					x = DogTag:ColorizeTag(x)
					return x:match("^|cff%x%x%x%x%x%x%[|r(.*)|cff%x%x%x%x%x%x%]|r$")
				end
			end
			return DogTag:ColorizeTag(text:sub(2, -2))
		end
		local function fix__handler(text)
			return _fix__handler(text):gsub("<", "&lt;"):gsub(">", "&gt;")
		end
		local function fix__handler2(text)
			return "|cffffffff" .. text:sub(1, -2) .. '|cffffffff"|r'
		end
		local function fix(text)
			return text:gsub('(%b"")', fix__handler2):gsub("(%b{})", fix__handler):gsub("&lbrace;", "{"):gsub("&rbrace;", "}")
		end
		
		local function dataToHTML(data)
			local t = newList()
			t[#t+1] = "<html>"
			t[#t+1] = "<body>"
			for _,h1 in ipairs(data) do
				local title = h1[1]
				t[#t+1] = "<h1>"
				t[#t+1] = fix(title)
				t[#t+1] = "</h1>"
				local description = type(h1[2]) == "string" and h1[2]
				if description then
					t[#t+1] = "<p>"
					t[#t+1] = fix(description)
					t[#t+1] = "</p>"
					t[#t+1] = "<br />"
				end
				for i = description and 3 or 2, #h1 do
					local h2 = h1[i]
					local title = h2[1]
					local description = type(h2[2]) == "string" and h2[2]
					t[#t+1] = "<h2>"
					t[#t+1] = fix(title)
					t[#t+1] = "</h2>"
					if description then
						t[#t+1] = "<p>"
						t[#t+1] = fix(description)
						t[#t+1] = "</p>"
						t[#t+1] = "<br />"
					end
					for j = description and 3 or 2, #h2 do
						local h3 = h2[j]
						local title = h3[1]
						t[#t+1] = "<h2>"
						t[#t+1] = fix(title)
						t[#t+1] = "</h2>"
						for k = 2, #h3 do
							local description = h3[k]
							t[#t+1] = "<p>"
							t[#t+1] = fix(description)
							t[#t+1] = "</p>"
							t[#t+1] = "<br />"
						end
					end
				end
			end
			while t[#t] == "<br />" do
				t[#t] = nil
			end
			t[#t+1] = "</body>"
			t[#t+1] = "</html>"
			local result = table.concat(t)
			t = del(t)
			return result
		end
		
		local syntaxHTML = dataToHTML({
			{
				"Syntax", [=[Syntax is in the standard form {alpha [Tag] bravo} where {alpha} is a literal word, {bravo} is a literal word and {[Tag]} will be replaced by the associated dynamic text. All tags and modifiers are case-insensitive, but will be corrected to proper casing if the tags are legal.]=],
				{ "Modifiers", [=[Modifiers can change how a tag's output looks. For example, the {{:HideZero}} modifier will hide the result if it is equal to the number {{0}}. e.g. {[CurHP:HideZero]} will show the current health except when it's equal to {{0}}, at which point it will be blank. You can chain together multiple modifiers as well, e.g. {[MissingHP:HideZero:Negate]} will show the missing health as a negative and not show it if it's equal to {{0}}.]=] },
				{ "Arguments", [=[Tags and modifiers can also take an argument, and can be fed in in a syntax similar to {[Tag(argument)]} or {[Tag:Modifier(argument)]}. You do not need to wrap in quotes for strings, this is done for you.]=] },
				{ "Logic Branching (if statements)", [=[
				* The {{?}} as used by DogTag serves as a boolean AND.<br />
				* The {{:}} as used by DogTag serves as a boolean OR.<br />
				* The {{!}} as used by DogTag serves as an else.
				]=],
					{ "Examples", [=[
						{[Status:SureHP:PercentHP]}<br />
						Will return one of the following (but only one): <br /><br />
						* "Dead", "Offline", "Ghost", etc -- no further information since the OR indicates that there is already a legitimate return<br />
						* "3560/8490" or "130/6575" (but not "62/100" unless the target in fact has {{100}} hit points) -- and not "0/2340" or "0/3592" because that would mean it is dead and that would have already been taken care of by the first tag in the sequence<br />
						* "25" or "35" or "72" (percent health) -- if the unit is not dead, offline, etc, and your addon is uncertain of your target's maximum and current health, it will display percent health.<br /><br />
						{[Status:[SureHP?CurHP]:PercentHP:Percent]} will deliver similar returns as to that above, but in a slightly different format which should be fairly apparent already.<br /><br />
						But to clarify, the nested {[SureHP?CurHP]} uses a boolean AND which means that if {{SureHP}} is nil or blank, the whole value is taken to be nil or blank. If {{SureHP}} is not nil or blank, the actual returned value of this nested expression is actually the term following the AND -- in this case, {{CurHP}}. So this will show {{CurHP}} if {{SureHP}} is found nonblank (if your addon knows how much health the unit has).
					]=], [=[
						{[IsFriend?MissingHP:Negate:Green!CurHP:Red]}<br /> 
						Will return one of the following (but only one):<br /><br />
						* If the unit is friendly, it will display the amount of health they must be healed to meet their maximum. It will be displayed in green, and with a negative sign in front of it.<br /> 
						* If the unit is enemy, it will display the current health. As this sequence is written, it will not consider whether it is a valid health value or not. On enemies where their health is uncertain, it will show a percentage (but not with a following percent sign), until a more reliable value can be determined by your addon. This value will be displayed in red.
					]=]}
				},
				{ "Unit specification", [=[
					Units are typically pre-specified by the addon which uses DogTag, whether {{#player}}, {{#mouseover}}, or otherwise. You can arbitrarily specify which unit a tag or modifier is interested in by using the form {[Tag#unit]} or {[Tag:Modifier#unit]}
				]=], { "Examples", [=[
					* {[CurHP#target]} - gets the current health of your target
					]=] },
					{ "List of example units", [=[
					* {{#player}} - your character<br />
					* {{#target}} - your target<br />
					* {{#targettarget}} - your target's target<br />
					* {{#pet}} - your pet<br />
					* {{#mouseover}} - the unit you are currently hovering over<br />
					* {{#focus}} - your focus unit<br />
					* {{#party1}} - the first member of your party<br />
					* {{#partypet2}} - the pet of the second member of your party<br />
					* {{#raid3}} - the third member of your raid<br />
					* {{#raidpet4}} - the pet of the fourth member of your raid<br />
					]=] }
				},
				{ "Negation of tags", [=[
				You can negate a tag or modifier by prefixing it with {{~}}. For tags, any non-blank value will turn blank and any blank value will turn into "True". For modifiers, any non-blank value will turn blank and any blank value will turn into the previous value (which in and of itself was non-blank). e.g. {[CurHP:~IsEqual(50)]} will hide any value not equal to 50 and {[~Offline]} will return "True" when online.
				]=]},
				{ "Internal expressions", [=[
				It is possible to embed DogTag sequences within other sequences. e.g. {[Text([Level] [Class])]} is functionally equivalent to {[Level] [Class]}. This can be useful in such occasions such as {[Level:IsLess([Level#player])]}. It is also possible to embed sequences within the core sequence itself, e.g. {[[Status:CurHP]:Green]} will highlight the result of {[Status:CurHP]} green, whereas {[Status:CurHP:Green]} would actually only highlight the result of {[CurHP]}.
				]=]},
				{ "Arithmetic operators", [=[
				You can use arithmetic operators in your DogTag sequences without issue, they function as expected with proper order-of-operations. {[1+2*3]} is functionally equivalent to {[1:Add([2:Mult(3)])]}<br /><br />
				* {{+}} - Addition<br />
				* {{-}} - Subtraction<br />
				* {{*}} - Multiplication<br />
				* {{/}} - Division<br />
				* {{%}} - Modulus<br />
				* {{^}} - Exponentiation
				]=]},
				{ "Comparison operators", [=[
				You can use comparison operators in your DogTag very similarly to arithmetic operators. {[1+2=3]} is functionally equivalent to {[1:Add([2]):IsEqual(3)]}<br /><br />
				* {{=}} - Equality<br />
				* {{~=}} - Inequality<br />
				* {{<}} - Less than<br />
				* {{>}} - Greater than<br />
				* {{<=}} - Less than or equal<br />
				* {{>=}} - Greater than or equal
				]=]}
			}
		})
		
		local function wrapWhite(text)
			return "|cffffffff" .. text .. "|r"
		end
		
		local function highlightWords(text)
			text = text:gsub("%f[A-Za-z0-9_](unit)%f[^A-Za-z0-9_]", wrapWhite)
			text = text:gsub("%f[A-Za-z0-9_](argument)%f[^A-Za-z0-9_]", wrapWhite)
			text = text:gsub("%f[A-Za-z0-9_](value)%f[^A-Za-z0-9_]", wrapWhite)
			text = text:gsub("%f[A-Za-z0-9_](number_value)%f[^A-Za-z0-9_]", wrapWhite)
			text = text:gsub("%f[A-Za-z0-9_](number)%f[^A-Za-z0-9_]", wrapWhite)
			return text
		end
		
		local tagsData = { 
			{"Tags", }
		}
		local tags = newList()
		local tagCategories_tmp = newSet(L["Miscellaneous"])
		for k,v in pairs(TagDocuments) do
			tags[#tags+1] = k
		end
		table.sort(tags)
		for k,v in pairs(TagDocumentCategories) do
			tagCategories_tmp[v] = true
		end
		local tagCategories = newList()
		for k in pairs(tagCategories_tmp) do
			tagCategories[#tagCategories+1] = k
		end
		tagCategories_tmp = del(tagCategories_tmp)
		table.sort(tagCategories)
		for _,category in ipairs(tagCategories) do
			local t = newList()
			for _,k in ipairs(tags) do
				if TagDocumentCategories[k] == category or (category == L["Miscellaneous"] and not TagDocumentCategories[k]) then
					local v = TagDocuments[k]
					local arg = TagArguments[k]
					if TagAliases[k] and not Tags[k] then
						-- alias
						local alias = TagAliases[k]
						if not alias:find("[%(:!%?%+%-%*/^%%<>=]") and not alias:find("~=") then
							local code, param, negate, unit = splitParam(alias)
							while TagAliases[code] do
								if TagAliases[code]:find("[%(:!%?%+%-%*/^%%<>=]") or TagAliases[code]:find("~=") then
									param = true
									break
								end
								local newParam, newNegate, newUnit
								code, newParam, newNegate, newUnit = splitParam(TagAliases[code])
								if not param then
									param = newParam
								end
								if newNegate then
									negate = not negate
								end
								if not unit then
									unit = newUnit
								end
							end
							if param then
								arg = nil
							else
								arg = TagArguments[code]
							end
						else
							arg = nil
						end
					end
					if not arg or arg:find("nil") then
						t[#t+1] = "{["
						t[#t+1] = k
						t[#t+1] = "]}"
					end
					if arg then
						if arg:find("nil") then
							t[#t+1] = " or "
						end
						t[#t+1] = "{["
						t[#t+1] = k
						t[#t+1] = "("
						if arg:find("string") then
							t[#t+1] = L["argument"]
						else -- number
							t[#t+1] = L["number"]
						end
						t[#t+1] = ")]}"
					end
					t[#t+1] = " - "
					t[#t+1] = highlightWords(v)
					if TagAliases[k] then
						t[#t+1] = " - "
						t[#t+1] = L["alias for "]
						t[#t+1] = "{["
						t[#t+1] = TagAliases[k]
						t[#t+1] = "]}"
					end
					t[#t+1] = "<br />"
					t[#t+1] = "e.g. "
					local examples = newList((";"):split(TagDocumentExamples[k]))
					for i, u in ipairs(examples) do
						if i > 1 then
							t[#t+1] = ", "
						end
						local tag, result = u:trim():match("(.*) => (.*)")
						t[#t+1] = "{"
						t[#t+1] = tag
						t[#t+1] = "}"
						t[#t+1] = " => "
						t[#t+1] = result
					end
					examples = del(examples)
					t[#t+1] = "<br />"
					t[#t+1] = "<br />"
				end
			end
			while t[#t] == "<br />" do
				t[#t] = nil
			end
			local result = table.concat(t)
			t = del(t)
			if result:len() > 0 then
				tagsData[1][#tagsData[1]+1] = { category, result }
			end
		end
		tagCategories = del(tagCategories)
		tags = del(tags)
		local tagsHTML = dataToHTML(tagsData)
		tagsData = nil
		
		local modifiersData = { 
			{"Modifiers", }
		}
		local modifiers = newList()
		for k in pairs(ModifierDocuments) do
			modifiers[#modifiers+1] = k
		end
			table.sort(modifiers)
		local modifierCategories_tmp = newSet(L["Miscellaneous"])
		for k,v in pairs(ModifierDocumentCategories) do
			modifierCategories_tmp[v] = true
		end
		local modifierCategories = newList()
		for k in pairs(modifierCategories_tmp) do
			modifierCategories[#modifierCategories+1] = k
		end
		modifierCategories_tmp = del(modifierCategories_tmp)
		table.sort(modifierCategories)
		for _,category in ipairs(modifierCategories) do
			local t = newList()
			for _,k in ipairs(modifiers) do
				if ModifierDocumentCategories[k] == category or (category == L["Miscellaneous"] and not ModifierDocumentCategories[k]) then
					local v = ModifierDocuments[k]
					local value = ModifierValues[k]
					local arg = ModifierArguments[k]
					local k_ = k
					if ModifierAliases[k] then
						-- alias
						local alias = ModifierAliases[k]
						local code, param, negate, unit = splitParam(alias)
						while ModifierAliases[code] do
							local newParam, newNegate, newUnit
							code, newParam, newNegate, newUnit = splitParam(ModifierAliases[code])
							if not param then
								param = newParam
							end
							if newNegate then
								negate = not negate
							end
							if not unit then
								unit = newUnit
							end
						end
						k_ = joinParam(code, param, negate, unit)
						value = ModifierValues[code]
						if param then
							arg = nil
						else
							arg = ModifierArguments[code]
						end
					end
					if not arg or arg:find("nil") then
						t[#t+1] = "{["
						if value == "number" then
							t[#t+1] = L["number_value"]
						else
							t[#t+1] = L["value"]
						end
						t[#t+1] = ":"
						t[#t+1] = k
						t[#t+1] = "]}"
					end
					if arg then
						if arg:find("nil") then
							t[#t+1] = " or "
						end
						t[#t+1] = "{["
						if value == "number" then
							t[#t+1] = L["number_value"]
						else
							t[#t+1] = L["value"]
						end
						t[#t+1] = ":"
						t[#t+1] = k
						t[#t+1] = "("
						if arg:find("string") then
							t[#t+1] = L["argument"]
						else -- number
							t[#t+1] = L["number"]
						end
						t[#t+1] = ")]}"
						
						if k_ == "Add" or k_ == "Sub" or k_ == "Mult" or k_ == "Div" or k_ == "Mod" or k_ == "Pow" then
							t[#t+1] = " or "
							t[#t+1] = "{["
							t[#t+1] = L["number_value"]
							if k_ == "Add" then
								t[#t+1] = "+"
							elseif k_ == "Sub" then
								t[#t+1] = "-"
							elseif k_ == "Mult" then
								t[#t+1] = "*"
							elseif k_ == "Div" then
								t[#t+1] = "/"
							elseif k_ == "Mod" then
								t[#t+1] = "%"
							elseif k_ == "Pow" then
								t[#t+1] = "^"
							end
							t[#t+1] = L["number"]
							t[#t+1] = "]}"
						elseif k_ == "IsEqual" or k_ == "~IsEqual" or k_ == "IsLess" or k_ == "IsLessEqual" or k_ == "~IsLessEqual" or k_ == "~IsLess" then
							t[#t+1] = " or "
							t[#t+1] = "{["
							t[#t+1] = L["value"]
							if k_ == "Add" then
							elseif k_ == "IsEqual" then
								t[#t+1] = "="
							elseif k_ == "~IsEqual" then
								t[#t+1] = "~="
							elseif k_ == "IsLess" then
								t[#t+1] = "<"
							elseif k_ == "IsLessEqual" then
								t[#t+1] = "<="
							elseif k_ == "~IsLessEqual" then
								t[#t+1] = ">"
							elseif k_ == "~IsLess" then
								t[#t+1] = ">="
							end
							t[#t+1] = L["argument"]
							t[#t+1] = "]}"
						end
					end
					t[#t+1] = " - "
					t[#t+1] = highlightWords(v)
					if ModifierAliases[k] then
						t[#t+1] = " - "
						t[#t+1] = L["alias for "]
						t[#t+1] = "{{:"
						t[#t+1] = ModifierAliases[k]
						t[#t+1] = "}}"
					end
					t[#t+1] = "<br />"
			
					t[#t+1] = "e.g. "
					local examples = newList((";"):split(ModifierDocumentExamples[k]))
					for i, u in ipairs(examples) do
						if i > 1 then
							t[#t+1] = ", "
						end
						local tag, result = u:trim():match("(.*) => (.*)")
						t[#t+1] = "{"
						t[#t+1] = tag
						t[#t+1] = "}"
						t[#t+1] = " => "
						t[#t+1] = result:gsub("{", "&lbrace;"):gsub("}", "&rbrace;"):gsub("<", "&lt;"):gsub(">", "&gt;")
					end
					examples = del(examples)
			
					t[#t+1] = "<br />"
					t[#t+1] = "<br />"
				end
			end
			
			while t[#t] == "<br />" do
				t[#t] = nil
			end
			local result = table.concat(t)
			t = del(t)
			if result:len() > 0 then
				modifiersData[1][#modifiersData[1]+1] = { category, result }
			end
		end
		modifiers = del(modifiers)
		modifierCategories = del(modifierCategories)
		local modifiersHTML = dataToHTML(modifiersData)
		modifiersData = nil
		
		local tabCount = 0
		local function makeTab(name)
			tabCount = tabCount + 1
			local tab = CreateFrame("Button", helpFrame:GetName() .. "Tab" .. tabCount, helpFrame, "TabButtonTemplate")
			tab:SetText(name)
			PanelTemplates_TabResize(0, tab)
			tab:SetID(tabCount)
			tab:SetScript("OnClick", function(this)
				PanelTemplates_SetTab(this:GetParent(), this:GetID())
				if this.Select then
					this:Select()
				end
				scrollBar:SetValue(0)
			end)
			tab:SetFrameLevel(tab:GetFrameLevel()+10)
			return tab
		end
		local tab1, tab2, tab3 = makeTab("Syntax"), makeTab("Tags"), makeTab("Modifiers")
		
		tab1:SetPoint("TOPLEFT", 30, -5)
		tab2:SetPoint("LEFT", tab1, "RIGHT")
		tab3:SetPoint("LEFT", tab2, "RIGHT")
		
		PanelTemplates_SetNumTabs(helpFrame, 3)
		helpFrame.selectedTab = 1
		PanelTemplates_SetTab(helpFrame, 1)
		
		function tab1:Select()
			html.text = syntaxHTML
			html:SetText(syntaxHTML)
		end
		
		function tab2:Select()
			html.text = tagsHTML
			html:SetText(tagsHTML)
		end
		
		function tab3:Select()
			html.text = modifiersHTML
			html:SetText(modifiersHTML)
		end
		
		tab1:Select()
	end
	function DogTag:OpenHelp()
		helpFrame:Show()
	end
end

function DogTag:AddFakeGlobal(key, value)
	assert(type(key) == "string")
	assert(type(value) == "table" or type(value) == "function")
	FakeGlobals["__" .. key] = value
end

function DogTag:AddAddonFinder(kind, name, func)
	assert(type(kind) == "string")
	assert(type(name) == "string")
	assert(type(func) == "function")
	AddonFinders[{kind, name, func}] = true
end

function DogTag:AddModifier(name, data)
	Modifiers[name] = data
end

function DogTag:AddTag(name, data)
	Tags[name] = data
end

function DogTag:AddTagAndModifier(name, data)
	self:AddTag(name, data)
	self:AddModifier(name, data)
end

DogTag_funcs[#DogTag_funcs+1] = function()
	_G.DogTag = nil
	DogTag.AddModifier = nil
	DogTag.AddTag = nil
	DogTag.AddTagAndModifier = nil
	DogTag.AddFakeGlobal = nil
	DogTag.AddAddonFinder = nil
	DogTag.Cleanup = nil
	DogTag.newList = nil
	DogTag.newDict = nil
	DogTag.newSet = nil
	DogTag.del = nil
	DogTag.IsLegitimateUnit = nil
	
	for tag, data in pairs(Tags) do
		if type(data) == "table" then
			if data.alias then
				if type(data.alias) == "string" then
					TagAliases[tag] = data.alias
				else -- function
					TagAliases[tag] = data.alias()
					TagAliasFunctions[tag] = data.alias
				end
			else
				local arg = sortStringList(data.arg)
				if arg and arg ~= "nil" then
					TagArguments[tag] = arg
					local a,b,c = (";"):split(arg)
					assert(a == "nil" or a == "number" or a == "string")
					assert(not b or b == "nil" or b == "number" or b == "string")
					assert(not c or c == "nil" or c == "number" or c == "string")
				end
				TagReturns[tag] = sortStringList(data.ret)
				if data.ret then
					local a,b,c = (";"):split(data.ret)
					assert(a == "nil" or a == "number" or a == "string")
					assert(not b or b == "nil" or b == "number" or b == "string")
					assert(not c or c == "nil" or c == "number" or c == "string")
				end
				TagEvents[tag] = sortStringList(data.events)
				TagGlobals[tag] = sortStringList(data.globals)
				if TagGlobals[tag] then
					local globals = newList((';'):split(TagGlobals[tag]))
					for _,v in ipairs(globals) do
						if not v:find("%.") then
							assert(_G[v], ("Unknown global: %q"):format(v))
						end
					end
					globals = del(globals)
				end
				TagAliases[tag] = data.fakeAlias
			end
			TagDocuments[tag] = data.doc
			TagDocumentExamples[tag] = data.example
			TagDocumentCategories[tag] = data.category
			local func = data[1]
			if Modifiers[tag] ~= Tags[tag] then
				del(Tags[tag])
			end
			if not data.alias then
				Tags[tag] = func
			else
				Tags[tag] = nil
			end
		elseif type(data) == "string" then
			TagAliases[tag] = data
			Tags[tag] = nil
		elseif type(data) == "function" then
			TagAliases[tag] = data()
			TagAliasFunctions[tag] = data
			Tags[tag] = nil
		end
	end

	for mod, data in pairs(Modifiers) do
		if type(data) == "table" then
			if data.alias then
				if type(data.alias) == "string" then
					ModifierAliases[mod] = data.alias
				else -- function
					ModifierAliases[mod] = data.alias()
					ModifierAliasFunctions[mod] = data.alias
				end
			else
				local arg = sortStringList(data.arg)
				if arg and arg ~= "nil" then
					ModifierArguments[mod] = arg
					local a,b,c = (";"):split(arg)
					assert(a == "nil" or a == "number" or a == "string")
					assert(not b or b == "nil" or b == "number" or b == "string")
					assert(not c or c == "nil" or c == "number" or c == "string")
				end
				ModifierReturns[mod] = sortStringList(data.ret)
				if data.ret then
					local a,b,c = (";"):split(data.ret)
					assert(a == "nil" or a == "number" or a == "string" or a == "same")
					assert(not b or b == "nil" or b == "number" or b == "string" or b == "same")
					assert(not c or c == "nil" or c == "number" or c == "string" or c == "same")
				end
				local value = sortStringList(data.value)
				if value and value ~= "number;string" then
					ModifierValues[mod] = value
					local a,b = (";"):split(value)
					assert(a == "number" or a == "string")
					assert(not b or b == "number" or b == "string")
				end
				ModifierEvents[mod] = sortStringList(data.events)
				ModifierGlobals[mod] = sortStringList(data.globals)
				if ModifierGlobals[mod] then
					local globals = newList((';'):split(ModifierGlobals[mod]))
					for _,v in ipairs(globals) do
						if not v:find("%.") then
							assert(_G[v], ("Unknown global: %q"):format(v))
						end
					end
					globals = del(globals)
				end
				ModifierAliases[mod] = data.fakeAlias
			end
			ModifierDocuments[mod] = data.doc
			ModifierDocumentExamples[mod] = data.example
			ModifierDocumentCategories[mod] = data.category
			local func = data[1]
			del(Modifiers[mod])
			if not data.alias then
				Modifiers[mod] = func
			else
				Modifiers[mod] = nil
			end
		elseif type(data) == "string" then
			ModifierAliases[mod] = data
			Modifiers[mod] = nil
		elseif type(data) == "function" then
			ModifierAliases[mod] = data()
			ModifierAliasFunctions[mod] = data
			Modifiers[mod] = nil
		end
	end
end

DogTag:AddAddonFinder("AceLibrary", "AceEvent-2.0", function(AceEvent)
	AceEvent:embed(DogTag)
	DogTag:RegisterEvent("Threat_ThreatUpdated")
end)
DogTag:AddAddonFinder("AceLibrary", "Babble-Spell-2.2", function(v) BabbleSpell = v end)
DogTag:AddAddonFinder("AceLibrary", "Threat-1.0", function(v) Threat = v end)

end